import bz2
import gzip
import json
import sys
import tempfile
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from csv import DictReader
from glob import glob
from io import BufferedReader, BytesIO, IOBase, TextIOWrapper
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
)

import pandas as pd
from httpx import AsyncClient
from yaml import safe_load

from ...model import JsonLikeDocument
from ...pluggable import Pluggable
from ...subclass_registry import MissingFromRegistryError, SubclassRegistry
from .credential_utils import AwsClientFactory
from .extractor import Extractor

SUPPORTED_FILE_FORMAT_REGISTRY = SubclassRegistry()
SUPPORTED_COMPRESSED_FILE_FORMAT_REGISTRY = SubclassRegistry()
SUPPORTED_FILE_SOURCES_REGISTRY = SubclassRegistry()


class Utf8TextIOWrapper(TextIOWrapper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, encoding="utf-8")


@abstractmethod
class ReadableFile:
    def as_reader(self, cls: type[IOBase]) -> AsyncContextManager[IOBase]:
        """Return a reader for the file.

        Given a class that represents a file reader, return an async context
        manager that yields an instance of that class. This allows different
        file types to be read in a uniform way.

        The context manager will close when the file is done being read and
        is no longer needed by the pipeline. Therefore, you can perform cleanup
        operations in the context manager on exit
        (i.e after the yield statement).
        """
        raise NotImplementedError

    def path_like(self) -> Path:
        """Return a Path object that represents the path to the file.

        This method should return a Path object that represents the path to the
        file. Note, the path does not need to be an actual file on the
        filesystem. It can be a URL or any other path-like object that can
        be used to identify the file.

        This will be called to sniff the file format and compression format of
        the file from the suffixes of the path.
        """
        raise NotImplementedError

    @asynccontextmanager
    async def popped_suffix_tempfile(
        self,
    ) -> AsyncContextManager[Tuple[Path, tempfile.NamedTemporaryFile]]:
        """Create a temporary file with the same suffixes sans the last one.

        This method creates a temporary file with the same suffixes as the
        original file sans the last one. This is useful when creating
        intermediary files that need to be passed through multiple codecs.

        For instance, imagine a file such as foo.txt.gz. When the file is
        being ingested, we will pass the file through the decompression codec
        and then through the file format codec. The decompression codec will
        remove the .gz suffix. And store it in an intermediary file which can
        be read by the file format codec (.txt in this case).
        """
        new = self.path_like().with_suffix("")
        with tempfile.NamedTemporaryFile(suffix="".join(new.suffixes)) as fp:
            yield Path(fp.name), fp


class LocalFile(ReadableFile):
    """A class that represents a local file that can be read.

    This class is used to read files from the local filesystem. The class
    takes a Path object that represents the path to the file. The class
    uses the Path object to open the file and return a reader that can be
    used to read the file.
    """

    def __init__(self, path: Path, pointer=None) -> None:
        self.path = path
        self.pointer = pointer

    @asynccontextmanager
    async def as_reader(self, cls: type[IOBase]):
        if self.pointer:
            yield cls(self.pointer)
        else:
            with self.path.open("rb") as fp:
                yield cls(fp)

    def path_like(self) -> Path:
        return self.path


class RemoteFile(ReadableFile):
    """A class that represents a remote file that can be read.

    This class is used to read files from a remote URL. The class uses the
    httpx library to stream the file from the URL and return a reader that
    can be used to read the file.
    """

    def __init__(
        self, url: str, client: AsyncClient, memory_spooling_in_mb: int
    ) -> None:
        self.url = url
        self.client = client
        self.max_memory_spooling = memory_spooling_in_mb * 1024 * 1024

    @asynccontextmanager
    async def as_reader(self, cls: type[IOBase]):
        if sys.version_info >= (3, 11):
            temp_file = tempfile.SpooledTemporaryFile(max_size=self.max_memory_spooling)
        else:
            temp_file = tempfile.TemporaryFile()

        with temp_file as fp:
            async with self.client.stream("GET", self.url) as response:
                async for chunk in response.aiter_bytes():
                    fp.write(chunk)
            fp.seek(0)
            yield cls(fp)

    def path_like(self) -> Path:
        return Path(self.url)


@SUPPORTED_FILE_SOURCES_REGISTRY.connect_baseclass
class FileSource(Pluggable, ABC):
    entrypoint_name = "file_sources"

    @classmethod
    def from_file_data_with_type_label(cls, data: Dict[str, Any]) -> "FileSource":
        type = data.pop("type")
        return SUPPORTED_FILE_SOURCES_REGISTRY.get(type).from_file_data(**data)

    @classmethod
    def from_file_data(cls, **kwargs) -> "FileSource":
        return cls(**kwargs)

    @abstractmethod
    def get_files(self) -> AsyncIterator[ReadableFile]:
        """Return an async iterator of files to be processed.

        This method should return an async iterator that yields instances of
        ReadableFile. The ReadableFile instances should be able to be read
        by the Extractor class. This method is used to abstract away the
        details of how files are read from different sources (i.e local files,
        remote files, etc).

        The async iterator should yield all the files that need to be processed
        by the pipeline. The Extractor class will then read the files and
        extract the records from them.
        """
        raise NotImplementedError

    def describe(self) -> str:
        """Return a human-readable description of the file source.

        This method should return a human-readable description of the file source.
        The description should be a string that describes the file source in a
        way that is understandable to the user. The description should be
        concise and informative.
        """
        return str(self)


@SUPPORTED_FILE_FORMAT_REGISTRY.connect_baseclass
class FileCodec(Pluggable, ABC):
    """Base class for file codecs.

    This class is used to define the interface for file codecs. File codecs
    are used to read files of a specific format. The class should implement
    the read_file_from_handle method which reads the file and returns an
    iterable of records.

    A class can define the `reader` attribute to specify the class that should
    be used to read the file. If the `reader` attribute is not defined, the
    default class used to read the file is a simple base io reader. In other
    words, if you don't define the `reader` attribute, the file will be read
    as a binary file. If you want to read the file as a text file, you should
    define the `reader` attribute as TextIOWrapper.
    """

    reader = None
    entrypoint_name = "file_formats"

    @abstractmethod
    def read_file_from_handle(self, reader: IOBase) -> Iterable[JsonLikeDocument]:
        """Read the file from the reader and return an iterable of records.

        This method should read the file from the reader and return an iterable
        of records. The records should be in a format that can be processed by
        the pipeline and be "JSON-like" (i.e a dictionary, list, etc).
        The method should yield the records one by one as they are read from
        the file.
        """
        raise NotImplementedError


@SUPPORTED_COMPRESSED_FILE_FORMAT_REGISTRY.connect_baseclass
class CompressionCodec(Pluggable, ABC):
    """Base class for compression codecs.

    This class is used to define the interface for compression codecs.
    Compression codecs are used to decompress files that are compressed
    using a specific compression algorithm. The class should implement the
    decompress_file method which decompresses the file and returns a new file
    that can be read by the file format codec or another decompression codec.
    """

    def __init__(self, file: ReadableFile) -> None:
        self.file = file

    @abstractmethod
    def decompress_file(self) -> AsyncContextManager[ReadableFile]:
        """Decompress the file and return a new file that can be read.

        This method should decompress the file and return a new file that can
        be read by the file format codec or another decompression codec.
        The method should return a new instance of ReadableFile that can be
        read by the pipeline.

        `ReadableFile` contains a helper method `popped_suffix_tempfile` that
        can be used to create a new file with the same suffixes as the original
        file sans the last suffix. This is useful when creating intermediary
        files that need to be passed through multiple codecs.

        For instance, imagine a file such as foo.txt.gz. When the file is
        being ingested, we will pass the file through the decompression codec
        and then through the file format codec. The decompression codec will
        remove the .gz suffix. And store it in an intermediary file which can
        be read by the file format codec (.txt in this case).
        """
        raise NotImplementedError


class JsonFileFormat(FileCodec, alias=".json"):
    """File format codec for JSON files.

    This class is used to read JSON files. The class reads the file and
    returns an iterable of records. The records are read one by one from
    the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".json". This means that when the pipeline encounters a file with the
    suffix .json, it will use this class to read the file.
    """

    reader = Utf8TextIOWrapper

    def read_file_from_handle(
        self, reader: TextIOWrapper
    ) -> Iterable[JsonLikeDocument]:
        return [json.load(reader)]


class LineSeperatedJsonFileFormat(FileCodec, alias=".jsonl"):
    """File format codec for line separated JSON files.

    This class is used to read line separated JSON files. The class reads the
    file and returns an iterable of records. The records are read one by one
    from the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".jsonl". This means that when the pipeline encounters a file with the
    suffix .jsonl, it will use this class to read the file.

    The class reads the file line by line and loads each line as a JSON object.
    """

    reader = Utf8TextIOWrapper

    def read_file_from_handle(
        self, reader: TextIOWrapper
    ) -> Iterable[JsonLikeDocument]:
        return (json.loads(line.strip()) for line in reader.readlines())


class ParquetFileFormat(FileCodec, alias=".parquet"):
    """File format codec for Parquet files.

    This class is used to read Parquet files. The class reads the file and
    returns an iterable of records. The records are read one by one from
    the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".parquet". This means that when the pipeline encounters a file with the
    suffix .parquet, it will use this class to read the file.

    The class reads the file using the pandas library and yields the records
    one by one as they are read.
    """

    def read_file_from_handle(self, fp: IOBase) -> Iterable[JsonLikeDocument]:
        df = pd.read_parquet(fp, engine="pyarrow")
        return (row[1].to_dict() for row in df.iterrows())


class TextFileFormat(FileCodec, alias=".txt"):
    """File format codec for text files.

    This class is used to read text files. The class reads the file and
    returns an iterable of records. The records are read one by one from
    the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".txt". This means that when the pipeline encounters a file with the
    suffix .txt, it will use this class to read the file.

    The class reads the file line by line and yields each line as a record in
    a dictionary with a single key "line".
    """

    reader = Utf8TextIOWrapper

    def read_file_from_handle(
        self, reader: TextIOWrapper
    ) -> Iterable[JsonLikeDocument]:
        return ({"line": line.strip()} for line in reader.readlines())


class CommaSeperatedValuesFileFormat(FileCodec, alias=".csv"):
    """File format codec for CSV files.

    This class is used to read CSV files. The class reads the file and
    returns an iterable of records. The records are read one by one from
    the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".csv". This means that when the pipeline encounters a file with the
    suffix .csv, it will use this class to read the file.

    The class reads the file using the csv.DictReader class and yields the
    records one by one as they are read
    """

    reader = Utf8TextIOWrapper

    def read_file_from_handle(
        self, reader: TextIOWrapper
    ) -> Iterable[JsonLikeDocument]:
        return DictReader(reader)


class YamlFileFormat(FileCodec, alias=".yaml"):
    """File format codec for YAML files.

    This class is used to read YAML files. The class reads the file and
    returns an iterable of records. The records are read one by one from
    the file and yielded as they are read.

    The class is registered with the file format registry using the alias
    ".yaml". This means that when the pipeline encounters a file with the
    suffix .yaml, it will use this class to read the file.

    The class reads the file using the yaml.safe_load function and yields one
    record which is the entire file.
    """

    reader = Utf8TextIOWrapper

    def read_file_from_handle(
        self, reader: TextIOWrapper
    ) -> Iterable[JsonLikeDocument]:
        return [safe_load(reader)]


class GzipFileFormat(CompressionCodec, alias=".gz"):
    """Compression codec for Gzip files.

    This class is used to decompress Gzip files. The class takes a file that
    is compressed using the Gzip compression algorithm and decompresses it.
    The class returns a new file that can be read by the file format codec
    or another decompression codec.

    The class is registered with the compression codec registry using the
    alias ".gz". This means that when the pipeline encounters a file with
    the suffix .gz, it will use this class to decompress the file.
    """

    @asynccontextmanager
    async def decompress_file(self) -> ReadableFile:
        async with self.file.popped_suffix_tempfile() as (new_path, temp_file):
            async with self.file.as_reader(BufferedReader) as reader:
                with gzip.GzipFile(fileobj=reader) as decompressor:
                    temp_file.write(decompressor.read())
            temp_file.seek(0)
            yield LocalFile(new_path, temp_file)


class Bz2FileFormat(CompressionCodec, alias=".bz2"):
    """Compression codec for BZ2 files.

    This class is used to decompress BZ2 files. The class takes a file that
    is compressed using the BZ2 compression algorithm and decompresses it.
    The class returns a new file that can be read by the file format codec
    or another decompression codec.

    The class is registered with the compression codec registry using the
    alias ".bz2". This means that when the pipeline encounters a file with
    the suffix .bz2, it will use this class to decompress the file.
    """

    @asynccontextmanager
    async def decompress_file(self) -> ReadableFile:
        async with self.file.popped_suffix_tempfile() as (new_path, temp_file):
            async with self.file.as_reader(BufferedReader) as reader:
                decompressor = bz2.BZ2Decompressor()
                for chunk in iter(lambda: reader.read(1024 * 1024), b""):
                    temp_file.write(decompressor.decompress(chunk))
            temp_file.seek(0)
            yield LocalFile(new_path, temp_file)


class LocalFileSource(FileSource, alias="local"):
    """A class that represents a source of local files to be read.

    This class is used to read files from the local filesystem. The class
    takes a list of glob patterns that are used to find files on the local
    filesystem. The class uses the glob module to find all the files that
    match the glob patterns and then yields instances of LocalFile that
    can be read by the pipeline.
    """

    @classmethod
    def from_file_data(cls, globs: List[str]) -> "LocalFileSource":
        all_matches = (
            Path(file)
            for glob_string in globs
            for file in glob(glob_string, recursive=True)
        )
        return cls(list(sorted(filter(Path.is_file, all_matches))))

    def __init__(self, paths: List[Path]) -> None:
        self.paths = paths

    async def get_files(self) -> AsyncIterator[ReadableFile]:
        for path in self.paths:
            yield LocalFile(path)

    def describe(self) -> str:
        if len(self.paths) == 1:
            return f"{self.paths[0]}"
        else:
            return f"{len(self.paths)} local files"


class RemoteFileSource(FileSource, alias="http"):
    """A class that represents a source of remote files to be read.

    This class is used to read files from remote URLs. The class takes a list
    of URLs and a memory spooling maximum size in MB. The class uses the httpx
    library to stream the files from the URLs and yield instances of RemoteFile
    that can be read by the pipeline.
    """

    def __init__(
        self, urls: Iterable[str], memory_spooling_max_size_in_mb: int = 10
    ) -> None:
        self.urls = urls
        self.memory_spooling_max_size = memory_spooling_max_size_in_mb * 1024 * 1024

    async def get_files(self) -> AsyncIterator[ReadableFile]:
        async with AsyncClient() as client:
            for url in self.urls:
                yield RemoteFile(url, client, self.memory_spooling_max_size)

    def describe(self) -> str:
        if len(self.urls) == 1:
            return f"{self.urls[0]}"
        else:
            return f"{len(self.urls)} remote files"


class S3File(ReadableFile):
    """A readable file that is stored in S3.

    This class is used to read files from S3. The class takes a key, bucket,
    and an S3 client. The class uses the S3 client to get the object from S3
    and yield an instance of the file that can be read by the pipeline.

    The class also has a method to archive the file after it has been read.

    """

    def __init__(
        self,
        key: str,
        s3_client,
        bucket: str,
        archive_dir: str | None,
        object_format: str | None,
    ) -> None:
        self.logger = getLogger(__name__)
        self.key = key
        self.s3_client = s3_client
        self.bucket = bucket
        self.archive_dir = archive_dir
        self.object_format = object_format

    def archive_if_required(self, key: str):
        if not self.archive_dir:
            return

        self.logger.info("Archiving S3 Object", extra=dict(key=key))
        filename = Path(key).name
        self.s3_client.copy(
            Bucket=self.bucket,
            Key=f"{self.archive_dir}/{filename}",
            CopySource={"Bucket": self.bucket, "Key": key},
        )
        self.s3_client.delete_object(Bucket=self.bucket, Key=key)

    def path_like(self) -> Path:
        path = Path(self.key)
        return path.with_suffix(self.object_format or path.suffix)

    @asynccontextmanager
    async def as_reader(self, reader: IOBase):
        streaming_body = self.s3_client.get_object(Bucket=self.bucket, Key=self.key)[
            "Body"
        ]
        yield reader(BytesIO(streaming_body.read()))
        self.archive_if_required(self.key)


class S3FileSource(FileSource, alias="s3"):
    """A class that represents a source of files stored in S3.

    This class is used to read files from S3. The class takes a bucket, prefix,
    and an S3 client. The class uses the S3 client to list the objects in the
    bucket and yield instances of S3File that can be read by the pipeline.

    The class also has a method to archive the file after it has been read.
    """

    @classmethod
    def from_file_data(
        cls,
        bucket: str,
        prefix: Optional[str] = None,
        archive_dir: Optional[str] = None,
        object_format: Optional[str] = None,
        **aws_client_args,
    ):
        return cls(
            bucket=bucket,
            prefix=prefix,
            archive_dir=archive_dir,
            object_format=object_format,
            s3_client=AwsClientFactory(**aws_client_args).make_client("s3"),
        )

    def __init__(
        self,
        bucket: str,
        s3_client,
        archive_dir: Optional[str] = None,
        object_format: Optional[str] = None,
        prefix: Optional[str] = None,
    ):
        self.bucket = bucket
        self.s3_client = s3_client
        self.archive_dir = archive_dir
        self.object_format = object_format
        self.prefix = prefix or ""

    def object_is_in_archive(self, key: str) -> bool:
        return key.startswith(self.archive_dir) if self.archive_dir else False

    def find_keys_in_bucket(self) -> Iterable[str]:
        # Returns all keys in the bucket that are not in the archive dir
        # and have the prefix.
        paginator = self.s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=self.bucket, Prefix=self.prefix)
        for page in page_iterator:
            keys = (obj["Key"] for obj in page.get("Contents", []))
            yield from filter(lambda k: not self.object_is_in_archive(k), keys)

    async def get_files(self):
        for key in self.find_keys_in_bucket():
            yield S3File(
                key=key,
                s3_client=self.s3_client,
                bucket=self.bucket,
                archive_dir=self.archive_dir,
                object_format=self.object_format,
            )


class FileExtractor(Extractor):
    """A class that extracts records from files.

    This class is used to extract records from files. The class takes a list
    of file sources that are used to read the files. The class uses the file
    sources to get a list of files and then reads the files using the file
    codecs. The class yields the records one by one as they are read from the
    files.
    """

    @classmethod
    def local(cls, globs: Iterable[str]):
        return FileExtractor.from_file_data([{"type": "local", "globs": globs}])

    @classmethod
    def s3(cls, **kwargs):
        return cls([S3FileSource.from_file_data(**kwargs)])

    @classmethod
    def remote(
        cls,
        urls: Iterable[str],
        memory_spooling_max_size_in_mb: int = 10,
    ):
        return FileExtractor.from_file_data(
            [
                {
                    "type": "http",
                    "urls": urls,
                    "memory_spooling_max_size_in_mb": memory_spooling_max_size_in_mb,
                }
            ]
        )

    @classmethod
    def from_file_data(cls, sources: List[Dict[str, Any]]) -> "FileExtractor":
        return cls(
            [FileSource.from_file_data_with_type_label(source) for source in sources]
        )

    def __init__(self, file_sources: Iterable[FileSource]) -> None:
        self.file_sources = file_sources
        self.logger = getLogger(__name__)

    async def read_file(self, file: ReadableFile) -> Iterable[JsonLikeDocument]:
        intermediaries: List[AsyncContextManager[ReadableFile]] = []

        while True:
            suffix = file.path_like().suffix

            # Try to find a compression codec that can decompress the file.
            # If a compression codec is found, decompress the file and store
            # the decompressed file in an intermediary file. The intermediary
            # file will be read by the file format codec.
            try:
                algo = SUPPORTED_COMPRESSED_FILE_FORMAT_REGISTRY.get(suffix)(file)
                decompression = algo.decompress_file()
                file = await decompression.__aenter__()
                intermediaries.append(decompression)
                continue
            except MissingFromRegistryError:
                pass

            # If we didn't find a compression codec, try to find a file format
            # codec that can read the file. If a file format codec is found,
            # read the file and yield the records.
            try:
                codec = SUPPORTED_FILE_FORMAT_REGISTRY.get(suffix)()
                async with file.as_reader(codec.reader or BufferedReader) as reader:
                    for record in codec.read_file_from_handle(reader):
                        yield record
            except MissingFromRegistryError:
                # If we didn't find a file format codec, break out of the loop
                # and yield no records.
                pass

            # Regardless of whether we found a codec or not, break out of the
            # loop and yield no more records becaause either (a) we found a
            # codec and read the file or (b) we didn't find a codec and
            # couldn't read the file. Trying again would be futile.
            break

        # Cleanup the intermediary files.
        for intermediary in reversed(intermediaries):
            await intermediary.__aexit__(None, None, None)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for file_source in self.file_sources:
            total_files_from_source = 0
            async for file in file_source.get_files():
                total_files_from_source += 1
                async for record in self.read_file(file):
                    yield record

            if total_files_from_source == 0:
                self.logger.warning(
                    f"No files found for source: {file_source.describe()}"
                )



from typing import Any, Dict, List, Optional

from ...model import NodeCreationRule, PropertySet
from ...pipeline.normalizers import LowercaseStrings
from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema import GraphObjectSchema, SchemaExpansionCoordinator
from .interpretation import Interpretation
from .property_mapping import PropertyMapping

# By default, data gathered from this interpretation is lower cased when a string.
DEFAULT_NORMALIZATION_ARGUMENTS = {LowercaseStrings.argument_flag(): True}


class SourceNodeInterpretation(Interpretation, alias="source_node"):
    """Stores information regarding the source node.

    Within a Pipeline File, a simple usage may look like this:

    ```yaml
    interpretations:
      # Conventionally, this should be first.
      - type: source_node
        node_type: TheTypeOfTheNodeYouAreIngesting
        key:
          id: !!python/jmespath value_for_id_field
    ```

    However, more complex usages may look like this to populate a rich node.

    ```yaml
    interpretations:
      - type: source_node
        node_type: Person
        key:
          first_name: !jmespath first_name
          last_name: !jmespath last_name
        properties:
           bio: !jmespath biography
    ```

    You may also apply additional labels and field level indexes:

    ```yaml
    interpretations:
      - type: source_node
        node_type: Person
        key:
          first_name: !jmespath first_name
          last_name: !jmespath last_name
        properties:
           phone_number: !jmespath phone
        additionally_index:
            - phone_number
        additional_types:
            - Customer
    ```

    You are also allowed by default to send normalization flags in:

    ```yaml
    interpretations:
      - type: source_node
        node_type: DomainName
        key:
          name: !jmespath name
        normalization:
          drop_trailing_dots: true
    ```
    """

    SOURCE_NODE_TYPE_ALIAS = "source_node"
    assigns_source_nodes = True

    __slots__ = (
        "node_type",
        "key",
        "properties",
        "additional_indexes",
        "additional_types",
        "norm_args",
        "allow_create",
        "update_last_ingested",
    )

    def __init__(
        self,
        node_type: StaticValueOrValueProvider,
        key: Dict[str, StaticValueOrValueProvider],
        properties: Optional[Dict[str, StaticValueOrValueProvider]] = None,
        additional_indexes: Optional[List[str]] = None,
        additional_types: Optional[List[str]] = None,
        normalization: Optional[Dict[str, Any]] = None,
        properties_normalization: Optional[Dict[str, Any]] = None,
        key_normalization: Optional[Dict[str, Any]] = None,
        allow_create: bool = True,
        update_last_ingested: bool = True,
    ):
        if normalization and (properties_normalization or key_normalization):
            raise ValueError(
                "You cannot specify normalization both at the root and at the key/properties level."
            )

        properties_normalization = properties_normalization or normalization
        key_normalization = key_normalization or normalization
        self.key_normalization = {
            **DEFAULT_NORMALIZATION_ARGUMENTS,
            **(key_normalization or {}),
        }
        self.properties_normalization = properties_normalization or {}
        self.node_type = ValueProvider.guarantee_value_provider(node_type)
        self.key = PropertyMapping.from_file_data(key or {})
        self.properties = PropertyMapping.from_file_data(properties or {})
        self.additional_indexes = additional_indexes or []
        self.additional_types = tuple(additional_types or [])
        if allow_create:
            self.creation_rule = NodeCreationRule.EAGER
        else:
            self.creation_rule = NodeCreationRule.MATCH_ONLY
        self.update_last_ingested = update_last_ingested

    def interpret(self, context: ProviderContext):
        normalized_key: PropertySet = PropertySet()
        self.key.apply_to(context, normalized_key, self.key_normalization)
        normalized_properties: PropertySet = PropertySet()
        self.properties.apply_to(
            context, normalized_properties, self.properties_normalization
        )

        context.desired_ingest.add_source_node(
            self.node_type.single_value(context),
            self.additional_types,
            self.creation_rule,
            normalized_key,
            normalized_properties,
        )

        if not self.update_last_ingested:
            context.desired_ingest.source.properties.remove_last_ingested()

    def expand_source_node_schema(self, source_node_schema: GraphObjectSchema):
        source_node_schema.add_keys(self.key)
        source_node_schema.add_properties(self.properties)
        source_node_schema.add_indexes(self.additional_indexes)
        source_node_schema.add_indexed_timestamp()

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        if not self.node_type.is_static:
            return

        coordinator.on_node_schema(
            self.expand_source_node_schema,
            alias=self.SOURCE_NODE_TYPE_ALIAS,
            node_type=self.node_type.value,
        )



import base64
import hashlib
import hmac
import pickle
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, TypeVar

from botocore.exceptions import ClientError

from ..pluggable import Pluggable
from ..subclass_registry import SubclassRegistry
from .extractors.credential_utils import AwsClientFactory

OBJECT_STORE_REGISTRY = SubclassRegistry(ignore_overrides=True)
T = TypeVar("T")


class InvalidSignatureError(ValueError):
    """An error that is raised when a signature is invalid."""

    def __init__(self, expected: bytes, actual: bytes):
        expected = base64.b64encode(expected).decode("utf-8")
        actual = base64.b64encode(actual).decode("utf-8")
        super().__init__(
            f"Invalid signature. Expected: {expected}, Actual: {actual}",
        )


class MalformedSignedObjectError(ValueError):
    """An error that is raised when a signed object is malformed."""

    def __init__(self):
        super().__init__("Malformed signed object.")


@dataclass(frozen=True, slots=True)
class SignedObject:
    """A signed object is an object that has been signed by a signer."""

    signature: bytes
    data: bytes

    def into_bytes(self) -> bytes:
        """Convert the signed object to a bytestring."""
        return self.signature + b"\n" + self.data

    @staticmethod
    def from_bytes(data: bytes) -> "SignedObject":
        """Create a signed object from a bytestring."""
        try:
            signature, data = data.split(b"\n", 1)
            return SignedObject(signature, data)
        except ValueError:
            raise MalformedSignedObjectError


class Signer(ABC):
    """A signer is used to sign and verify objects in an object store."""

    @staticmethod
    def hmac(base_64_secret: str) -> "Signer":
        """Create a HMAC signature from a base64 secret."""
        return HmacSigner.from_base64(base_64_secret)

    @abstractmethod
    def sign(self, data: bytes) -> SignedObject:
        """Sign an object.

        This method should sign the given object and return a signed object
        that contains the signature and the data. The signature should be
        a bytestring.
        """
        pass

    @abstractmethod
    def verify(self, signed: SignedObject):
        """Verify a signed object.

        This method should verify the given signed object. If the signature is
        invalid, this should raise an InvalidSignatureError. If the signature
        is valid, this should do nothing.
        """
        pass


class HmacSigner(Signer):
    """A signer that uses HMAC to sign objects.

    This signer uses HMAC with SHA-256 to sign objects. The key used for
    signing is provided when the signer is created. The key should be a
    bytestring. The signer will use the key to sign objects using HMAC with
    SHA-256.
    """

    def __init__(self, key: bytes):
        self.key = key

    @staticmethod
    def from_base64(key: str):
        """Create the signer from a base64-encoded key."""
        return HmacSigner(base64.b64decode(key))

    def _get_digest(self, data: bytes) -> bytes:
        return hmac.new(self.key, data, hashlib.sha256).digest()

    def sign(self, data: bytes) -> SignedObject:
        return SignedObject(self._get_digest(data), data)

    def verify(self, signed: SignedObject):
        expected = self._get_digest(signed.data)
        if not hmac.compare_digest(signed.signature, expected):
            raise InvalidSignatureError(expected, signed.signature)


class Namespace(ABC):
    """A namespace is a way to scope keys in an object store."""

    def scope(self, key: str) -> str:
        """Scope a key.

        This method should return a new key that is scoped by the namespace.
        The new key should be unique for the given key and namespace.
        """
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class StaticNamespace(Namespace):
    """A namespace that is static and does not change."""

    prefix: str

    def scope(self, key: str) -> str:
        return f"{self.prefix}/{key}"


@OBJECT_STORE_REGISTRY.connect_baseclass
class ObjectStore(ABC, Pluggable):
    entrypoint_name = "object_stores"

    @abstractmethod
    def get(self, key: str) -> Optional[bytes]:
        """Get an object from the object store.

        The object store should return a path to the object that can be read.
        If an error is encountered, this should raise an exception.
        If the object is not found, this should return None.
        """
        pass

    @abstractmethod
    def put(self, key: str, path: bytes):
        """Put an object into the object store.

        The object store should store the object at the given path under the
        given key. If an error is encountered, this should raise an exception.
        """
        pass

    @abstractmethod
    def delete(self, key: str):
        """Delete an object from the object store.

        If the object is not found, this should do nothing.
        """
        pass

    def get_pickled(self, key: str) -> T:
        """Get an object from the object store that has been pickled."""
        if (data := self.get(key)) is None:
            return None

        return pickle.loads(data)

    def put_picklable(self, key: str, value: T):
        """Put an object from the object store that has can be pickled."""
        self.put(key, pickle.dumps(value))

    def namespaced(self, namespace: Namespace | str) -> "ObjectStore":
        """Return a new object store with a namespace applied.

        This method should return a new object store that is the same as the
        current object store, but with the given namespace applied. The new
        object store should be able to retrieve and store objects with keys
        that are scoped by the namespace.

        Note that `namespaced` can be called multiple times to apply multiple
        namespaces to the same object store. The namespaces will be applied in
        a stack-like manner, with the most recent namespace being the
        outermost.

        Args:
            namespace: The namespace to apply to the object store.
        """
        if isinstance(namespace, str):
            namespace = StaticNamespace(namespace)
        return NamespacedObjectStore(self, namespace)

    def signed(self, signer: "Signer") -> "ObjectStore":
        """Return a new object store with a signer applied.

        This method should return a new object store that is the same as the
        current object store, but with the given signer applied. The new object
        store should be able to retrieve and store signed objects.

        Args:
            signer: The signer to apply to the object store.
        """
        return SignedObjectStore(self, signer)

    @staticmethod
    def null() -> "ObjectStore":
        return NullObjectStore()

    @staticmethod
    def from_file_arguments(type: str, **arguments) -> "ObjectStore":
        object_store_type = OBJECT_STORE_REGISTRY.get(type)
        return object_store_type(**arguments)


class DirectoryObjectStore(ObjectStore, alias="local"):
    """An object store that stores objects in a directory on a file system."""

    def __init__(self, root: Optional[Path] = None) -> None:
        self.root = root or (Path.cwd() / ".nodestream" / "objects")

    def get(self, key: str) -> Optional[bytes]:
        path = self.root / key
        if not path.exists():
            return None

        with path.open("rb") as f:
            return f.read()

    def put(self, key: str, path: bytes):
        target = self.root / key
        target.parent.mkdir(parents=True, exist_ok=True)
        with target.open("wb") as f:
            f.write(path)

    def delete(self, key: str):
        target = self.root / key
        target.unlink(missing_ok=True)


class NullObjectStore(ObjectStore, alias="null"):
    """An object store that does not store any objects."""

    def get(self, _: str) -> Optional[bytes]:
        return None

    def put(self, _path: str, _data: bytes):
        pass

    def delete(self, _path: str):
        pass


@dataclass(frozen=True, slots=True)
class NamespacedObjectStore(ObjectStore):
    """An object store that stores objects with a namespace applied.

    This object store wraps another object store and scopes keys with a
    provided namespace. This is used for ensuring that keys are sufficiently
    unique when they are stored and read from the object store.
    """

    store: ObjectStore
    namespace: Namespace

    def get(self, key: str) -> Optional[bytes]:
        return self.store.get(self.namespace.scope(key))

    def put(self, key: str, path: bytes):
        self.store.put(self.namespace.scope(key), path)

    def delete(self, key: str):
        self.store.delete(self.namespace.scope(key))


@dataclass(frozen=True, slots=True)
class SignedObjectStore(ObjectStore):
    """An object store that stores signed objects.

    This object store wraps another object store and signs and verifies objects
    that are stored and retrieved from that object store. This is used for
    ensuring that objects are not tampered with when they are stored and read
    from the object store.
    """

    store: ObjectStore
    signer: Signer

    def get(self, key: str) -> Optional[bytes]:
        if (data := self.store.get(key)) is None:
            return None

        signed = SignedObject.from_bytes(data)
        self.signer.verify(signed)
        return signed.data

    def put(self, key: str, path: bytes):
        signed = self.signer.sign(path)
        self.store.put(key, signed.into_bytes())

    def delete(self, key: str):
        self.store.delete(key)


class S3ObjectStore(ObjectStore, alias="s3"):
    __slots__ = ("client", "bucket_name")

    def __init__(self, bucket_name: str, **client_factory_args):
        client_factory = AwsClientFactory(**client_factory_args)
        self.client = client_factory.make_client("s3")
        self.bucket_name = bucket_name

    def get(self, key: str) -> Optional[bytes]:
        try:
            response = self.client.get_object(Bucket=self.bucket_name, Key=key)
            return response["Body"].read()
        except ClientError as e:
            status = e.response["ResponseMetadata"]["HTTPStatusCode"]
            if status == 404:
                return None
            raise e

    def put(self, key: str, data: bytes):
        self.client.put_object(Bucket=self.bucket_name, Key=key, Body=data)

    def delete(self, key: str):
        try:
            self.client.delete_object(Bucket=self.bucket_name, Key=key)
        except ClientError as e:
            status = e.response["ResponseMetadata"]["HTTPStatusCode"]
            if status == 404:
                return
            raise e



import asyncio
from dataclasses import asdict
from logging import getLogger

from ..model import (
    IngestionHookRunRequest,
    Node,
    NodeCreationRule,
    RelationshipWithNodes,
    TimeToLiveConfiguration,
)
from .ingest_strategy import IngestionStrategy
from .operation_debouncer import OperationDebouncer
from .query_executor import QueryExecutor


class DebouncedIngestStrategy(IngestionStrategy, alias="debounced"):
    def __init__(self, query_executor: QueryExecutor) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.executor = query_executor
        self.debouncer = OperationDebouncer()
        self.hooks_saved_for_after_ingest = []

    async def ingest_source_node(
        self, source: Node, creation_rule: NodeCreationRule = NodeCreationRule.EAGER
    ) -> None:
        self.debouncer.debounce_node_operation(source, node_creation_rule=creation_rule)

    async def ingest_relationship(self, relationship: RelationshipWithNodes):
        self.debouncer.debounce_node_operation(
            relationship.from_node, relationship.from_side_node_creation_rule
        )
        self.debouncer.debounce_node_operation(
            relationship.to_node, relationship.to_side_node_creation_rule
        )
        self.debouncer.debounce_relationship(relationship)

    async def run_hook(self, request: IngestionHookRunRequest):
        if request.before_ingest:
            await self.executor.execute_hook(request.hook)
        else:
            self.hooks_saved_for_after_ingest.append(request.hook)

    async def perform_ttl_operation(self, config: TimeToLiveConfiguration):
        self.logger.debug("Executing TTL", extra=asdict(config))
        await self.executor.perform_ttl_op(config)
        self.logger.info("Executed TTL", extra=asdict(config))

    def flush_nodes_updates(self):
        update_coroutines = (
            self.executor.upsert_nodes_in_bulk_with_same_operation(
                operation, node_group
            )
            for operation, node_group in self.debouncer.drain_node_groups()
        )
        return asyncio.gather(*update_coroutines)

    async def flush_relationship_updates(self):
        # Because databases tend to require exclusive locks on both nodes,
        # and these updates are very likely to be related to at least one of the nodes
        # in the relationship, we need to update relationships one operation type
        # at a time to avoid deadlocks.
        for rel_shape, rel_group in self.debouncer.drain_relationship_groups():
            await self.executor.upsert_relationships_in_bulk_of_same_operation(
                rel_shape, rel_group
            )

    async def flush(self):
        await self.flush_nodes_updates()
        await self.flush_relationship_updates()

        # Because we don't know what exactly the hooks do, we can't reliably parallelize
        # them because we could overwhelm the database so we are going to do them one at
        # a time.
        for hook in self.hooks_saved_for_after_ingest:
            await self.executor.execute_hook(hook)
        self.hooks_saved_for_after_ingest.clear()

    async def finish(self):
        """Close connector by calling finish method from Step"""
        await self.executor.finish()



from typing import Iterable

from ..metrics import Metric, Metrics
from ..model import IngestionHook, Node, RelationshipWithNodes, TimeToLiveConfiguration
from .query_executor import (
    OperationOnNodeIdentity,
    OperationOnRelationshipIdentity,
    QueryExecutor,
)


class QueryExecutorWithStatistics(QueryExecutor):
    __slots__ = ("inner",)

    def __init__(self, inner: QueryExecutor) -> None:
        self.inner = inner

    async def upsert_nodes_in_bulk_with_same_operation(
        self, operation: OperationOnNodeIdentity, nodes: Iterable[Node]
    ):
        await self.inner.upsert_nodes_in_bulk_with_same_operation(operation, nodes)
        Metrics.get().increment(Metric.NODES_UPSERTED, len(nodes))

    async def upsert_relationships_in_bulk_of_same_operation(
        self,
        shape: OperationOnRelationshipIdentity,
        relationships: Iterable[RelationshipWithNodes],
    ):
        await self.inner.upsert_relationships_in_bulk_of_same_operation(
            shape, relationships
        )
        Metrics.get().increment(Metric.RELATIONSHIPS_UPSERTED, len(relationships))

    async def perform_ttl_op(self, config: TimeToLiveConfiguration):
        await self.inner.perform_ttl_op(config)
        Metrics.get().increment(Metric.TIME_TO_LIVE_OPERATIONS)

    async def execute_hook(self, hook: IngestionHook):
        await self.inner.execute_hook(hook)
        Metrics.get().increment(Metric.INGEST_HOOKS_EXECUTED)

    async def finish(self):
        await self.inner.finish()



from .auto_change_detector import AutoChangeDetector, MigratorInput
from .auto_migration_maker import AutoMigrationMaker
from .migrations import Migration, MigrationGraph
from .migrator import (
    Migrator,
    OperationTypeNotSupportedError,
    OperationTypeRoutingMixin,
)
from .project_migrations import ProjectMigrations

__all__ = (
    "AutoChangeDetector",
    "AutoMigrationMaker",
    "Migrator",
    "MigratorInput",
    "Migration",
    "MigrationGraph",
    "ProjectMigrations",
    "OperationTypeRoutingMixin",
    "OperationTypeNotSupportedError",
)



import time
from abc import ABC
from dataclasses import dataclass, field
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Optional, Tuple

from pandas import Timestamp

from .creation_rules import NodeCreationRule, RelationshipCreationRule

if TYPE_CHECKING:
    from .desired_ingestion import DesiredIngestion


class DeduplicatableObject(ABC):
    def get_dedup_key(self) -> tuple:
        raise NotImplementedError

    def update(self, other: "DeduplicatableObject"):
        raise NotImplementedError


# This is a weird little trick to cache the current time for a short period of
# time. When nodestream is adding the timestamp to tones of node and
# relationship objects, it can be a bit slow. Just to get times that are
# basically the same, there is little value in getting the current time over
# and over again. So we cache the current time for a short period of time.
# Profiling this generally shows a 5-15% speedup in the time it takes to
# ingest a large number of nodes and relationships (Excluding database time).


@lru_cache(maxsize=128)
def _get_cached_timestamp(time_hash: int) -> Timestamp:
    del time_hash  # Unused; just needed to cache the result
    return Timestamp.utcnow()


def get_cached_timestamp(
    max_age_in_seconds: int = 2, epoch: Optional[float] = None
) -> Timestamp:
    # calling time.time() is relatively fast, but getting a Timestamp object
    # is relatively slow.
    epoch = epoch or time.time()
    ttl_hash = round(epoch) // max_age_in_seconds
    return _get_cached_timestamp(ttl_hash)


class PropertySet(dict):
    def set_property(self, property_key: str, property_value: Any):
        self[property_key] = property_value

    @classmethod
    def default_properties(cls) -> "PropertySet":
        from ..metrics import Metrics

        """Returns a default set of properties which set values.

        These default values indicate when the current pipeline touched the object the properties are for.
        """
        pipeline_name = Metrics.get().pipeline_name
        now = get_cached_timestamp()
        return cls(
            {
                "last_ingested_at": now,
                f"last_ingested_by_{pipeline_name}_at": now,
                f"was_ingested_by_{pipeline_name}": True,
            }
        )

    @classmethod
    def empty(cls) -> "PropertySet":
        """Returns an empty property set."""
        return PropertySet()

    def apply(self, key_value_gen):
        if not key_value_gen:
            return

        for key, val in key_value_gen:
            self.set_property(key, val)

    def merge(self, properties: "PropertySet"):
        if not properties:
            return

        for key, val in properties.items():
            self.set_property(key, val)

    def remove_last_ingested(self):
        self.pop("last_ingested_at", None)


@dataclass(slots=True)
class Node(DeduplicatableObject):
    """A `Node` is an entity that has a distinct identity.

    Each `Node` represents an entity (a person, place, thing, category or other piece of data) that has a distinct
    identity. Nodestream assumes the underlying graph database layer is a Labeled Property Graph. The identity
    of a node is defined by a root `type` (sometimes referred to as a label) as well as set of property name, value pairs
    representing the primary key of that node. In a relational database, this would be the combination of the table name
    as well as the primary key columns.

    The node class also can store additional property key value pairs that are not considered part of the identity of the
    node but rather additional data. In a relational database, these would be the non-primary key columns.

    A `Node` can also store additional types (labels) that can apply additional categorization in the database. This has
    no direct analogy in a relational database.
    """

    type: Optional[str] = None
    key_values: PropertySet = field(default_factory=PropertySet.empty)
    properties: PropertySet = field(default_factory=PropertySet.default_properties)
    additional_types: Tuple[str] = field(default_factory=tuple)

    @property
    def has_valid_id(self) -> bool:
        # Return that some of the ID values are defined.
        all_key_parts_defined = all(
            value is not None for value in self.key_values.values()
        )
        return len(self.key_values) > 0 and all_key_parts_defined

    @property
    def is_valid(self) -> bool:
        return self.has_valid_id and self.type is not None

    @property
    def identity_shape(self) -> "NodeIdentityShape":
        return NodeIdentityShape(
            type=self.type,
            keys=tuple(self.key_values.keys()),
            additional_types=self.additional_types,
        )

    def has_same_key(self, other: "Node") -> bool:
        return self.key_values == other.key_values

    def update(self, other: "Relationship"):
        self.properties.update(other.properties)

    def get_dedup_key(self) -> tuple:
        return tuple(sorted(self.key_values.items()))

    def into_ingest(self) -> "DesiredIngestion":
        from .desired_ingestion import DesiredIngestion

        return DesiredIngestion(source=self)


@dataclass(slots=True, frozen=True)
class Relationship(DeduplicatableObject):
    """A `Relationship` represents an inherent connection between two `Node`s.

    Each `Relationship` follows a relatively similar model to a `Node`. There is a _single_ type for the relationship.
    Relationships can store properties on the relationship itself (This would be similar to a jump table in a relational database).

    A key for a `Relationship` can also be provided. By default, `nodestream` will assume that there should be one
    relationship of the same type between two nodes. By providing keys, `nodestream` will create multiple relationships between
    two nodes and will discriminate based on the key values.

    This model represents the relationship itself and DOES NOT include a reference of the nodes that are stored.
    """

    type: str
    key_values: PropertySet = field(default_factory=PropertySet.empty)
    properties: PropertySet = field(default_factory=PropertySet.default_properties)

    @property
    def identity_shape(self) -> "RelationshipIdentityShape":
        return RelationshipIdentityShape(
            type=self.type, keys=tuple(self.key_values.keys())
        )

    def has_same_key(self, other: "Node") -> bool:
        return self.key_values == other.key_values

    def update(self, other: "Relationship"):
        self.properties.update(other.properties)

    def get_dedup_key(self) -> tuple:
        return tuple(sorted(self.key_values.values()))


@dataclass(slots=True)
class RelationshipWithNodes(DeduplicatableObject):
    """Stores information about the related node and the relationship itself."""

    from_node: Node
    to_node: Node
    relationship: Relationship
    outbound: bool = True  # lets us maintain knowledge of which node is the source node

    to_side_node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER
    from_side_node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER
    relationship_creation_rule: RelationshipCreationRule = (
        RelationshipCreationRule.EAGER
    )

    def has_same_keys(self, other: "RelationshipWithNodes") -> bool:
        return (
            self.to_node.has_same_key(other.to_node)
            and self.from_node.has_same_key(other.from_node)
            and self.relationship.has_same_key(other.relationship)
        )

    def update(self, other: "RelationshipWithNodes"):
        self.to_node.properties.update(other.to_node.properties)
        self.from_node.properties.update(other.from_node.properties)
        self.relationship.properties.update(other.relationship.properties)

    def get_dedup_key(self) -> tuple:
        return (
            self.to_node.get_dedup_key(),
            self.from_node.get_dedup_key(),
            self.relationship.get_dedup_key(),
        )

    def into_ingest(self) -> "DesiredIngestion":
        from .desired_ingestion import DesiredIngestion

        ingest = DesiredIngestion(source=self.from_node)
        ingest.add_relationship(
            self.to_node,
            self.relationship,
            outbound=True,
            node_creation_rule=NodeCreationRule.MATCH_ONLY,
            relationship_creation_rule=RelationshipCreationRule.CREATE,
        )
        return ingest


@dataclass(slots=True, frozen=True)
class NodeIdentityShape:
    type: str
    keys: Tuple[str]
    additional_types: Tuple[str] = field(default_factory=tuple)


@dataclass(slots=True, frozen=True)
class RelationshipIdentityShape:
    type: str
    keys: Tuple[str]



from asyncio import create_task, gather
from logging import getLogger
from typing import Iterable, List, Tuple

from ..metrics import Metric, Metrics
from ..schema import ExpandsSchema, ExpandsSchemaFromChildren
from .channel import StepInput, StepOutput, channel
from .object_storage import ObjectStore
from .progress_reporter import PipelineProgressReporter
from .step import Step, StepContext


class StepExecutor:
    """`StepExecutor` is a utility that is used to run a step in a pipeline.

    The `StepExecutor` is responsible for starting, stopping, and running a
    step in a pipeline. It is used to execute a step by passing records
    between the input and output channels of the step.
    """

    __slots__ = ("step", "input", "output", "context")

    def __init__(
        self,
        step: Step,
        input: StepInput,
        output: StepOutput,
        context: StepContext,
    ) -> None:
        self.step = step
        self.input = input
        self.output = output
        self.context = context

    async def start_step(self):
        try:
            Metrics.get().increment(Metric.STEPS_RUNNING)
            await self.step.start(self.context)
        except Exception as e:
            self.context.report_error("Error starting step", e)

    async def stop_step(self):
        try:
            Metrics.get().decrement(Metric.STEPS_RUNNING)
            await self.step.finish(self.context)
        except Exception as e:
            self.context.report_error("Error stopping step", e)

    async def emit_record(self, record):
        can_continue = await self.output.put(record)
        if not can_continue:
            self.context.debug(
                "Downstream is not accepting more records. Gracefully stopping."
            )

        return can_continue

    async def drive_step(self):
        try:
            while (next_record := await self.input.get()) is not None:
                results = self.step.process_record(next_record, self.context)
                async for record in results:
                    if not await self.emit_record(record):
                        return

            async for record in self.step.emit_outstanding_records(self.context):
                if not await self.emit_record(record):
                    return

            self.context.debug("Step finished emitting")
        except Exception as e:
            self.context.report_error("Error running step", e, fatal=True)

    async def run(self):
        self.context.debug("Starting step")
        await self.start_step()
        await self.drive_step()
        await self.output.done()
        self.input.done()
        await self.stop_step()
        self.context.debug("Finished step")


class PipelineOutput:
    """`PipelineOutput` is an output channel for a pipeline.

    A `PipelineOutput` is used to consume records from the last step in a
    pipeline and report the progress of the pipeline.
    """

    __slots__ = ("input", "reporter")

    def __init__(self, input: StepInput, reporter: PipelineProgressReporter):
        self.input = input
        self.reporter = reporter

    def call_handling_errors(self, f, *args):
        try:
            f(*args)
        except Exception:
            self.reporter.logger.exception(f"Error running {f.__name__}")

    async def run(self):
        """Run the pipeline output.

        This method is used to run the pipeline output. It will consume records
        from the last step in the pipeline and report the progress of the
        pipeline using the `PipelineProgressReporter`. The pipeline output will
        block until all records have been consumed from the last step in the
        pipeline.
        """
        metrics = Metrics.get()
        self.call_handling_errors(self.reporter.on_start_callback)

        index = 0
        while (obj := await self.input.get()) is not None:
            metrics.increment(Metric.RECORDS)
            self.call_handling_errors(self.reporter.report, index, obj)
            index += 1

        self.call_handling_errors(self.reporter.on_finish_callback, metrics)


class Pipeline(ExpandsSchemaFromChildren):
    """`Pipeline` is a collection of steps that are executed in sequence.

    A `Pipeline` is a collection of steps that are executed in sequence. Each
    step processes records and emits new records that are passed to the next
    step in the pipeline. The pipeline is responsible for starting, stopping,
    and running the steps in the pipeline.
    """

    __slots__ = ("steps", "step_outbox_size", "logger", "object_store")

    def __init__(
        self,
        steps: Tuple[Step, ...],
        step_outbox_size: int,
        object_store: ObjectStore,
    ) -> None:
        self.steps = steps
        self.step_outbox_size = step_outbox_size
        self.logger = getLogger(self.__class__.__name__)
        self.object_store = object_store

    def get_child_expanders(self) -> Iterable[ExpandsSchema]:
        return (s for s in self.steps if isinstance(s, ExpandsSchema))

    async def run(self, reporter: PipelineProgressReporter):
        """Run the pipeline.

        This method is used to run the pipeline. It will start, stop, and run
        the steps in the pipeline in sequence. It will pass records between the
        steps in the pipeline using channels. The pipeline will report on the
        progress of the pipeline using the `PipelineProgressReporter`. The
        pipeline will run in an asynchronous context. The pipeline will block
        until all steps in the pipeline are finished.

        This method does not return anything. If an error occurs during the
        processing of the pipeline, it will be reported using the
        `PipelineProgressReporter`.

        Args:
            channel_size: The size of the channels used to pass records between
                steps in the pipeline.
            reporter: The `PipelineProgressReporter` used to report on the
                progress of the pipeline.
        """
        # Create the input and output channels for the pipeline. The input
        # channel is used to pass records from the previous step to the current
        # step. The output channel is used to pass records from the current
        # step to the next step. The channels are used to pass records between
        # the steps in the pipeline. The channels have a fixed size to control
        # the flow of records between the steps.
        executors: List[StepExecutor] = []
        current_input, current_output = channel(self.step_outbox_size)
        pipeline_output = PipelineOutput(current_input, reporter)

        # Create the executors for the steps in the pipeline. The executors
        # will be used to run the steps concurrently. The steps are created in
        # reverse order so that the output of each step is connected to the
        # input of the next step.
        for reversed_index, step in reversed(list(enumerate(self.steps))):
            index = len(self.steps) - reversed_index - 1
            storage = self.object_store.namespaced(str(index))
            context = StepContext(step.__class__.__name__, index, reporter, storage)
            current_input, next_output = channel(self.step_outbox_size)
            exec = StepExecutor(step, current_input, current_output, context)
            current_output = next_output
            executors.append(exec)

        # There is a "leftover" input channel that is not connected to any
        # step. This channel is connected to the first step in the pipeline
        # so we can mark it as done since we are not going to produce anything
        # onto it.
        await current_output.done()

        # Run the pipeline by running all the steps and the pipeline output
        # concurrently. This will block until all steps are finished.
        running_steps = (create_task(executor.run()) for executor in executors)

        self.logger.info("Starting Pipeline")
        await gather(*running_steps, create_task(pipeline_output.run()))
        self.logger.info("Pipeline Completed")



from cleo.helpers import option

from ..operations import ExecuteMigrations
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION, TARGETS_OPTION


class RunMigrations(NodestreamCommand):
    name = "migrations run"
    description = "Execute pending migrations on the specified target."
    options = [
        PROJECT_FILE_OPTION,
        TARGETS_OPTION,
        option("all-targets", "a", "Run migrations on all targets", flag=True),
    ]

    async def handle_async(self):
        project = self.get_project()
        migrations = self.get_migrations()

        if self.option("all-targets"):
            targets = [target for target in project.targets_by_name]
        else:
            targets = self.option(TARGETS_OPTION.name)

        if len(targets) == 0:
            self.info("No targets specified, nothing to do.")
            return

        for target_name in targets:
            target = project.get_target_by_name(target_name)
            await self.run_operation(ExecuteMigrations(migrations, target))



from dataclasses import dataclass, field
from typing import Any, Dict

from ..file_io import LazyLoadedArgument
from ..schema.migrations import Migrator

WRITER_ARGUMENTS = ("batch_size", "collect_stats", "ingest_strategy_name")


@dataclass(slots=True, frozen=True)
class Target:
    name: str
    connector_config: Dict[str, Any]
    writer_arguments: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_file_data(cls, name: str, file_data):
        writer_arguments = {
            item: value
            for item in WRITER_ARGUMENTS
            if (value := file_data.pop(item, None)) is not None
        }

        return cls(
            name=name,
            connector_config=file_data,
            writer_arguments=writer_arguments,
        )

    @property
    def resolved_connector_config(self):
        return LazyLoadedArgument.resolve_if_needed(self.connector_config)

    @property
    def connector(self):
        from ..databases import DatabaseConnector

        return DatabaseConnector.from_database_args(**self.resolved_connector_config)

    def make_writer(self):
        from ..databases import GraphDatabaseWriter

        return GraphDatabaseWriter.from_connector(
            connector=self.connector, **self.writer_arguments
        )

    def make_type_retriever(self):
        return self.connector.make_type_retriever()

    def make_migrator(self) -> "Migrator":
        return self.connector.make_migrator()

    def to_file_data(self):
        return dict(**self.connector_config, **self.writer_arguments)



from cleo.helpers import option

from ...metrics import Metrics
from ..operations import (
    ExecuteMigrations,
    InitializeLogger,
    InitializeMetricsHandler,
    InitializeProject,
    RunPipeline,
)
from .nodestream_command import NodestreamCommand
from .shared_options import (
    JSON_OPTION,
    MANY_PIPELINES_ARGUMENT,
    PROJECT_FILE_OPTION,
    PROMETHEUS_OPTIONS,
    TARGETS_OPTION,
)


class Run(NodestreamCommand):
    name = "run"
    description = "run a pipeline in the current project"
    arguments = [MANY_PIPELINES_ARGUMENT]
    options = [
        PROJECT_FILE_OPTION,
        JSON_OPTION,
        TARGETS_OPTION,
        option(
            "annotations",
            "a",
            "An annotation to apply to the pipeline during initialization. Steps without one of these annotations will be skipped.",
            multiple=True,
            flag=False,
        ),
        option(
            "reporting-frequency",
            "r",
            "How often to report progress",
            default=1000,
            flag=False,
        ),
        option(
            "step-outbox-size",
            "s",
            "How many records to buffer in each step's outbox before blocking",
            default=1000,
            flag=False,
        ),
        option(
            "auto-migrate",
            description="Ensure all specified targets are migrated before running specified pipelines",
            flag=True,
        ),
        option(
            "storage-backend",
            description="Storage backend to use for checkpointing",
            flag=False,
        ),
        *PROMETHEUS_OPTIONS,
    ]

    async def auto_migrate_targets_if_needed(self, project):
        if not self.option("auto-migrate"):
            return

        targets = self.option("targets")
        migrations = self.get_migrations()
        for target_name in targets:
            target = project.get_target_by_name(target_name)
            await self.run_operation(ExecuteMigrations(migrations, target))

    async def handle_async(self):
        with Metrics.capture():
            await self.run_operation(InitializeLogger())
            await self.run_operation(InitializeMetricsHandler())
            project = await self.run_operation(InitializeProject())
            await self.auto_migrate_targets_if_needed(project)
            await self.run_operation(RunPipeline(project))



import asyncio
from asyncio import Queue, wait_for
from typing import Optional, Tuple

from ..metrics import Metric, Metrics


#  WARNING: DoneObject refernces should not exist beyond this module.
class DoneObject:
    """A `DoneObject` is a marker object that indicates that a step is done.

    When a step is done processing records, it should emit a `DoneObject` to
    indicate that it is finished. This allows downstream steps to know when a
    step is done and they can stop processing records.

    The `DoneObject` is a special object that is used to signal the end of a
    stream of records. It is not a record itself and should not be processed as
    such.
    """

    pass


CHANNEL_TIMEOUT = 0.25


class Channel:
    """`Channel` is a communication channel between steps in a pipeline.

    A `Channel` is used to pass records between steps in a pipeline. It is a
    simple queue that can be used to pass records from one step to another. The
    `Channel` is asynchronous and can be used to pass records between steps in
    an asynchronous context.

    A `Channel` has a fixed size and will block if the queue is full. This
    allows the pipeline to control the flow of records between steps and
    prevent one step from overwhelming another step with too many records.
    """

    __slots__ = ("queue", "input_dropped", "metrics")

    def __init__(self, size: int) -> None:
        self.queue = Queue(maxsize=size)
        self.input_dropped = False
        self.metrics = Metrics.get()

    async def get(self):
        """Get an object from the channel.

        This method is used to get an object from the channel. It will block
        until an object is available in the channel. If the channel is empty,
        it will block until an object is available.

        Returns:
            object: The object that was retrieved from the channel.
        """
        object = await self.queue.get()
        self.metrics.decrement(Metric.BUFFERED_RECORDS)
        return object

    async def put(self, obj) -> bool:
        """Put an object in the channel.

        This method is used to put an object in the channel. It will block
        until the object is put in the channel. If the channel is full, it
        will block until there is space in the channel.

        Returns:
            bool: True if the object was successfully put in the channel, False
            if the channel is full and the object was not put in the channel.
        """
        try:
            await wait_for(self.queue.put(obj), timeout=CHANNEL_TIMEOUT)
            self.metrics.increment(Metric.BUFFERED_RECORDS)
            return True
        except (TimeoutError, asyncio.TimeoutError):
            return False


class StepOutput:
    """`StepOutput` is an output channel for a step in a pipeline.

    A `StepOutput` is used to pass records from a step to the next step in a
    pipeline. It is a simple wrapper around a `Channel` that provides a more
    convenient interface for putting records in the channel.
    """

    __slots__ = ("channel",)

    def __init__(self, channel: Channel) -> None:
        self.channel = channel

    async def done(self):
        """Mark the output channel as done.

        This method is used to mark the output channel as done. It will put a
        `DoneObject` in the channel to indicate that the step is finished
        processing records. This allows downstream steps to know when the step
        is done and they can stop processing records.
        """
        await self.put(DoneObject)

    async def put(self, obj) -> bool:
        """Put an object in the output channel.

        This method is used to put an object in the output channel. It will
        block until the object is put in the channel. If the channel is full,
        it will block until there is space in the channel unless the channel is
        closed on the other end.

        Returns:
            bool: True if the object was successfully put in the channel, False
            if the channel is closed on the other end and the object was not
            put in the channel.
        """
        successfully_put = False
        while not successfully_put and not self.channel.input_dropped:
            successfully_put = await self.channel.put(obj)

        return successfully_put


class StepInput:
    """`StepInput` is an input channel for a step in a pipeline.

    A `StepInput` is used to get records from the previous step in a pipeline.
    It is a simple wrapper around a `Channel` that provides a more convenient
    interface for getting records from the channel.

    The `StepInput` is an asynchronous generator that can be used to get
    records from the channel in an asynchronous context.
    """

    __slots__ = ("channel",)

    def __init__(self, channel: Channel) -> None:
        self.channel = channel

    async def get(self) -> Optional[object]:
        """Get an object from the input channel.

        This method is used to get an object from the input channel. It will
        block until an object is available in the channel. Once the channel is
        closed (i.e. a `DoneObject` is received), this method will return None.

        Returns:
            object: The object that was retrieved from the channel.
        """
        if (object := await self.channel.get()) is DoneObject:
            return None

        return object

    def done(self):
        """Mark the input channel as done.

        This will close the channel.
        """
        self.channel.input_dropped = True


def channel(size: int) -> Tuple[StepInput, StepOutput]:
    """Create a new input and output channel.

    Args:
        size: The size of the channel.
    """
    channel = Channel(size)
    return StepInput(channel), StepOutput(channel)



from logging import getLogger
from typing import Any, AsyncGenerator, Dict, Optional

from boto3.dynamodb.types import TypeDeserializer

from ...credential_utils import AwsClientFactory
from ...extractor import Extractor


class DynamoRecord:
    """
    Collection of attributes and their data.
    """

    @classmethod
    def from_raw_dynamo_record(cls, raw_dynamo_entry: Dict[str, Dict[str, Any]]):
        return DynamoRecord(
            {
                attribute_name: TypeDeserializer().deserialize(attribute_value)
                for attribute_name, attribute_value in raw_dynamo_entry.items()
            }
        )

    def __init__(self, record_data: Dict[str, Any]):
        self.record_data = record_data


class DynamoDBExtractor(Extractor):
    @classmethod
    def from_file_data(
        cls,
        table_name: str,
        limit: int = 100,
        scan_filter: Dict[str, Any] = {},
        projection_expression: str = None,
        filter_expression: str = None,
        **aws_client_args,
    ):
        client = AwsClientFactory(**aws_client_args).make_client("dynamodb")
        return cls(
            client=client,
            table_name=table_name,
            limit=limit,
            scan_filter=scan_filter,
            projection_expression=projection_expression,
            filter_expression=filter_expression,
        )

    def __init__(
        self,
        client,
        table_name,
        limit: int,
        scan_filter: Dict[str, Any],
        projection_expression: Optional[str] = None,
        filter_expression: Optional[str] = None,
    ) -> None:
        self.client = client
        self.logger = getLogger(self.__class__.__name__)
        self.table_name = table_name
        self.limit = limit
        self.scan_filter = scan_filter
        self.projection_expression = projection_expression
        self.filter_expression = filter_expression
        self.last_evaluated_key = None

    @property
    def effective_parameters(self):
        tentative_parameters = {
            "TableName": self.table_name,
            "ScanFilter": self.scan_filter,
            "ProjectionExpression": self.projection_expression,
            "FilterExpression": self.filter_expression,
            "ExclusiveStartKey": self.last_evaluated_key,
            "PaginationConfig": {
                "PageSize": self.limit,
            },
        }

        return {key: value for key, value in tentative_parameters.items() if value}

    def scan_table(self):
        paginator = self.client.get_paginator("scan")
        pages = paginator.paginate(**self.effective_parameters)
        for page in pages:
            self.last_evaluated_key = page.get("LastEvaluatedKey")
            yield from page["Items"]

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for record in self.scan_table():
            yield DynamoRecord.from_raw_dynamo_record(record).record_data

    async def resume_from_checkpoint(self, checkpoint_object):
        self.last_evaluated_key = checkpoint_object.get("last_evaluated_key")

    async def make_checkpoint(self):
        return {"last_evaluated_key": self.last_evaluated_key}



from decimal import Decimal
from typing import Any, Iterable, Type

from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider, ValueProviderException


class CastValueProvider(ValueProvider):
    CONVERTERS = {
        "int": int,
        "float": float,
        "decimal": Decimal,
        "str": str,
        "bool": bool,
        "bytes": bytes,
    }

    def __init__(self, data: ValueProvider, to: str):
        self.data = data
        self.to = to

    def cast_value(self, value):
        return self.CONVERTERS[self.to](value)

    def single_value(self, context: ProviderContext) -> Any:
        try:
            return self.cast_value(self.data.single_value(context))
        except Exception as e:
            raise ValueProviderException(str(context.document), self) from e

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        try:
            return map(self.cast_value, self.data.many_values(context))
        except Exception as e:
            raise ValueProviderException(str(context.document), self) from e

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!cast", lambda loader, node: cls(**loader.construct_mapping(node))
        )

    def __str__(self):
        return f"CastValueProvider: { {'data': str(self.data), 'to': str(self.to)} }"


SafeDumper.add_representer(
    CastValueProvider,
    lambda dumper, cast: dumper.represent_mapping(
        "!cast",
        {"data": cast.data, "to": cast.to},
    ),
)



from contextlib import asynccontextmanager
from typing import List

from .migrations import Migration, MigrationGraph
from .operations import Operation


class Migrator:
    """Abstract class for executing migrations.

    This class is used by the schema change detector to execute migrations
    after they have been detected.
    """

    async def acquire_lock(self) -> None:
        """Acquire a lock against the database to prevent concurrent ops.

        This method should block until the lock is acquired. If the lock cannot
        be acquired, then this method should raise an exception.
        """
        pass

    async def release_lock(self) -> None:
        """Release the lock against the database.

        This method should release the lock against the database. If the lock
        cannot be released, then this method should raise an exception.
        """
        pass

    async def start_transaction(self) -> None:
        """Start a transaction.

        For databases that do not support transactions, this method should be a
        no-op. For databases that do support transactions, this method should
        start a transaction. If the transaction cannot be started, then this
        method should raise an exception.
        """
        pass

    async def commit_transaction(self) -> None:
        """Commit a transaction.

        For databases that do not support transactions, this method should be a
        no-op. For databases that do support transactions, this method should
        commit the transaction. If the transaction cannot be committed, then
        this method should raise an exception.

        This method is only called if start_transaction() was called and
        did not raise an exception.
        """
        pass

    async def rollback_transaction(self) -> None:
        """Rollback a transaction.

        For databases that do not support transactions, this method should be a
        no-op. For databases that do support transactions, this method should
        rollback the transaction. If the transaction cannot be rolled back,
        then this method should raise an exception.

        This method is only called if start_transaction() was called and did
        not raise an exception.
        """
        pass

    async def execute_operation(self, operation: Operation) -> None:
        """Execute an operation.

        Takes an operation and executes it against the database. If the
        operation cannot be executed, then this method should raise an
        exception. If the operation is executed, then this method should
        return None.

        This method is only called if start_transaction() was called and did
        not raise an exception.

        Args:
            operation: The operation to execute.
        """
        pass

    async def mark_migration_as_executed(self, migration: Migration) -> None:
        """Mark a migration as executed.

        This method should mark a migration as executed. If the migration
        cannot be marked as executed, then this method should raise an
        exception.

        This method is only called if start_transaction() was called and did
        not raise an exception.

        By default, this method does nothing. This method should be overridden
        by subclasses to mark a migration as executed assuming that the
        migrations are stateful.

        Args:
            migration: The migration to mark as executed.
        """
        pass

    async def get_completed_migrations(self, graph: MigrationGraph) -> List[Migration]:
        """Get a list of completed migrations.

        This method should return a list of completed migrations. If the
        migrations cannot be retrieved, then this method should raise an
        exception.

        By default, this method returns an empty list. This method should be
        overridden by subclasses to return a list of completed migrations
        assuming that the migrations are stateful.

        Returns:
            A list of completed migrations.
        """
        return []

    @asynccontextmanager
    async def transaction(self):
        await self.start_transaction()
        try:
            yield
            await self.commit_transaction()
        except Exception as e:
            await self.rollback_transaction()
            raise e

    async def execute_migration(self, migration: Migration) -> None:
        """Execute a migration.

        Args:
            migration: The migration to execute.
        """
        async with self.transaction():
            await self.acquire_lock()

        async with self.transaction():
            try:
                for operation in migration.operations:
                    await self.execute_operation(operation)
            except Exception:
                await self.release_lock()
                raise

        async with self.transaction():
            await self.mark_migration_as_executed(migration)

        async with self.transaction():
            await self.release_lock()


class OperationTypeNotSupportedError(Exception):
    """Raised when an operation type is not supported."""

    pass


class OperationTypeRoutingMixin:
    """Mixin for routing operations to the correct method.

    This mixin is used for routing operations to the correct method. It provides
    a default implementation of execute_operation() that routes the operation
    to the correct method based on the operation type.

    The method that an operation is routed to is determined by converting the
    operation type to snake case and prepending it with "execute_". For
    example, the operation type "CreateNodeType" would be routed to the method
    "execute_create_node_type()".

    If the operation type cannot be routed to a method, then an
    OperationTypeNotSupportedError is raised.

    The method that an operation is routed to must be implemented by the class
    that uses this mixin.
    """

    async def execute_operation(self, operation: Operation) -> None:
        type_as_snake = operation.type_as_snake_case()
        method_name = f"execute_{type_as_snake}"
        try:
            method = getattr(self, method_name)
            await method(operation)
        except AttributeError:
            raise OperationTypeNotSupportedError(
                f"Unable to route operation to method: {method_name}"
            )



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Iterable, List, Optional

from ...file_io import LoadsFromYamlFile, SavesToYamlFile
from .operations import Operation


@dataclass(frozen=True, slots=True)
class Migration(LoadsFromYamlFile, SavesToYamlFile):
    """A migration is a collection of operations that can be applied to a schema.

    Migrations are used to migrate the database schema from one version to another.
    """

    name: str
    operations: List[Operation]
    dependencies: List[str]
    replaces: List[str] = field(default_factory=list)

    def to_file_data(self):
        return {
            "name": self.name,
            "operations": [operation.to_file_data() for operation in self.operations],
            "dependencies": self.dependencies,
            "replaces": self.replaces,
        }

    @classmethod
    def from_file_data(cls, file_data):
        return cls(
            name=file_data["name"],
            operations=[
                Operation.from_file_data(operation_file_data)
                for operation_file_data in file_data["operations"]
            ],
            dependencies=file_data["dependencies"],
            replaces=file_data.get("replaces", []),
        )

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                "name": str,
                "operations": [Operation.describe_yaml_schema()],
                "dependencies": [str],
                Optional("replaces"): [str],
            }
        )

    def is_root_migration(self) -> bool:
        """Check if this migration is a root migration.

        Returns:
            True if this migration is a root migration, False otherwise.
        """
        return len(self.dependencies) == 0

    def is_leaf_migration(self, graph: "MigrationGraph") -> bool:
        """Check if this migration is a leaf migration.

        Returns:
            True if this migration is a leaf migration, False otherwise.
        """
        dependents = (m for m in graph.all_migrations() if self.name in m.dependencies)
        return not any(dependents)

    def write_to_file_with_default_name(self, directory: Path) -> Path:
        """Save the migration to a directory with a default name.

        Args:
            directory: The directory to save the migration to.
        """
        path = directory / f"{self.name}.yaml"
        self.write_to_file(path)
        return path

    def is_squashed_migration(self) -> bool:
        """Check if this migration is a squashed migration.

        Returns:
            True if this migration is a squashed migration, False otherwise.
        """
        return len(self.replaces) > 0

    @classmethod
    def squash(
        cls,
        new_name: str,
        migrations: Iterable["Migration"],
        optimize_operations: bool = True,
    ) -> "Migration":
        """Make a new migration as the squashed for the given migrations.

        Args:
            new_name: The name of the new migration.
            migrations: The migrations to squash.
            optimize_operations: Whether to optimize the operations
            before squashing.

        Returns:
            The new migration that is the effective squash of the
            given migrations.
        """
        names_of_migrations_being_squashed = {m.name for m in migrations}
        effective_operations = [o for m in migrations for o in m.operations]
        dependencies = list(
            {
                d
                for m in migrations
                for d in m.dependencies
                if d not in names_of_migrations_being_squashed
            }
        )
        if optimize_operations:
            effective_operations = Operation.optimize(effective_operations)

        return cls(
            name=new_name,
            operations=effective_operations,
            replaces=list(names_of_migrations_being_squashed),
            dependencies=dependencies,
        )


@dataclass(frozen=True, slots=True)
class MigrationGraph:
    """A directed acyclic graph of migrations.

    The graph is represented as an adjacency list.
    """

    migrations_by_name: Dict[str, Migration]

    def get_migration(self, name: str) -> Migration:
        """Get a migration by name.

        Args:
            name: The name of the migration to get.

        Returns:
            The migration with the given name.

        Raises:
            KeyError: If there is no migration with the given name.
        """
        return self.migrations_by_name[name]

    def get_ordered_migration_plan(
        self, completed_migrations: List[Migration]
    ) -> List[Migration]:
        completed_migration_names = {m.name for m in completed_migrations}
        replacement_index = {r: m for m in self.all_migrations() for r in m.replaces}
        plan_order = []

        for migration in self.topological_order():
            if migration.name in completed_migration_names:
                continue

            # If we are considering a migration that has been replaced,
            # we _only_ want to add it if at least one of the migrations
            # replaced by the replacing migration has not been completed.
            # IN otherwords, if the changes from a squashed migration have
            # have (at least partially) been applied, we don't want to add
            # the squashed migration to the plan but will want to add the
            # migrations that were replaced to the plan.
            if (replacement := replacement_index.get(migration.name)) is not None:
                if not any(
                    r in completed_migration_names for r in replacement.replaces
                ):
                    continue

            # Similarly, if we are looking at a squashed migration, we want to
            # add it to the plan only if none of the migrations that it replaces
            # have been completed.
            if migration.is_squashed_migration():
                if any(r in completed_migration_names for r in migration.replaces):
                    continue

            # Now here we are in one of three stats:
            #
            # 1. The migraiton is some "regular" migration that has not been
            #       completed.
            # 2. The migration is a squashed migration that we want to add to
            #       the plan.
            # 3. The migration is a migration that has been replaced by a
            #       squashed migration but we've determined that at least one
            #       of the migrations that it replaces has not been completed.
            #
            # In all of these cases, we want to add the migration to the plan.
            plan_order.append(migration)

        return plan_order

    def _iterative_dfs_traversal(self, *start_node: Migration) -> List[Migration]:
        visited_order = []
        visited_set = set()
        stack = [(n, False) for n in start_node]

        while stack:
            node, processed = stack.pop()
            if node.name in visited_set:
                continue
            if processed:
                visited_order.append(node)
                visited_set.add(node.name)
                continue

            stack.append((node, True))
            stack.extend(
                ((self.migrations_by_name[n], False) for n in node.dependencies)
            )

        return visited_order

    def squash_between(
        self,
        name: str,
        from_migration: Migration,
        to_migration: Optional[Migration] = None,
    ):
        """Squash all migrations between two migrations.

        Args:
            name: The name of the new squashed migration.
            from_migration: The migration to start squashing from.
            to_migration: The migration to stop squashing at.

        Returns:
            The new squashed migration.
        """
        ordered = self.topological_order()
        from_index = ordered.index(from_migration)
        to_index = ordered.index(to_migration) if to_migration else len(ordered) - 1
        migrations_to_squash = ordered[from_index : to_index + 1]
        return Migration.squash(name, migrations_to_squash)

    def topological_order(self):
        return self._iterative_dfs_traversal(*self.get_leaf_migrations())

    def all_migrations(self) -> Iterable[Migration]:
        """Iterate over all migrations in the graph."""
        return self.migrations_by_name.values()

    def get_leaf_migrations(self) -> Iterable[Migration]:
        """Iterate over all leaf migrations in the graph."""
        return [
            migration
            for migration in self.all_migrations()
            if migration.is_leaf_migration(self)
        ]

    @classmethod
    def from_directory(cls, directory: Path) -> "MigrationGraph":
        """Load a migration graph from a directory."""

        migrations = (
            Migration.read_from_file(migration_file)
            for migration_file in directory.glob("*.yaml")
        )
        return cls.from_iterable(migrations)

    @classmethod
    def from_iterable(cls, migrations: Iterable[Migration]) -> "MigrationGraph":
        """Load a migration graph from a list of migrations."""
        return cls(migrations_by_name={m.name: m for m in migrations})



from cleo.helpers import option

from ..operations import GenerateMigration
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION


class MakeMigration(NodestreamCommand):
    name = "migrations make"
    description = "Generate a migration for the current project."
    options = [
        PROJECT_FILE_OPTION,
        option(
            "dry-run",
            description="Don't generate the migration file, just output the detected changes.",
            flag=True,
        ),
    ]

    async def handle_async(self):
        project = self.get_project()
        schema = project.get_schema()
        migrations = self.get_migrations()
        dry_run = self.option("dry-run")
        await self.run_operation(GenerateMigration(migrations, schema, dry_run))



from abc import ABC, abstractmethod

from ..state import Schema
from .migrations import MigrationGraph
from .migrator import Migrator, OperationTypeRoutingMixin
from .operations import (
    AddAdditionalNodePropertyIndex,
    AddAdditionalRelationshipPropertyIndex,
    AddNodeProperty,
    AddRelationshipProperty,
    CreateNodeType,
    CreateRelationshipType,
    DropAdditionalNodePropertyIndex,
    DropAdditionalRelationshipPropertyIndex,
    DropNodeProperty,
    DropNodeType,
    DropRelationshipProperty,
    DropRelationshipType,
    NodeKeyExtended,
    NodeKeyPartRenamed,
    RelationshipKeyExtended,
    RelationshipKeyPartRenamed,
    RenameNodeProperty,
    RenameNodeType,
    RenameRelationshipProperty,
    RenameRelationshipType,
)


class StateProvider(ABC):
    @abstractmethod
    async def get_schema(self) -> Schema:
        pass


class StaticStateProvider(StateProvider):
    __slots__ = ("schema",)

    def __init__(self, schema: Schema) -> None:
        self.schema = schema

    async def get_schema(self) -> Schema:
        return self.schema


class MigrationGraphStateProvider(StateProvider):
    __slots__ = ("migrations",)

    def __init__(self, migrations: MigrationGraph) -> None:
        self.migrations = migrations

    async def get_schema(self) -> Schema:
        migrator = InMemoryMigrator()
        for migration in self.migrations.get_ordered_migration_plan([]):
            await migrator.execute_migration(migration)
        return migrator.schema


class InMemoryMigrator(OperationTypeRoutingMixin, Migrator):
    """A migrator that executes migrations in memory.

    This migrator executes migrations in memory. It does not actually execute
    migrations against a database. It is used for testing the schema change
    detector as well as for generating the current application state from the
    migrations in the migration graph.
    """

    __slots__ = ("schema",)

    def __init__(self) -> None:
        self.schema = Schema()

    async def execute_create_node_type(self, operation: CreateNodeType) -> None:
        self.schema.put_node_type(operation.as_node_type())

    async def execute_create_relationship_type(
        self, operation: CreateRelationshipType
    ) -> None:
        self.schema.put_relationship_type(operation.as_relationship_type())

    async def execute_drop_node_type(self, operation: DropNodeType) -> None:
        self.schema.drop_node_type_by_name(operation.name)

    async def execute_drop_relationship_type(
        self, operation: DropRelationshipType
    ) -> None:
        self.schema.drop_relationship_type_by_name(operation.name)

    async def execute_rename_node_property(self, operation: RenameNodeProperty) -> None:
        node_type_def = self.schema.get_node_type_by_name(operation.node_type)
        node_type_def.rename_property(
            operation.old_property_name, operation.new_property_name
        )

    async def execute_rename_relationship_property(
        self, operation: RenameRelationshipProperty
    ) -> None:
        relationship_type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        relationship_type_def.rename_property(
            operation.old_property_name, operation.new_property_name
        )

    async def execute_rename_node_type(self, operation: RenameNodeType) -> None:
        current_type = self.schema.get_node_type_by_name(operation.old_type)
        self.schema.drop_node_type_by_name(operation.old_type)
        current_type.name = operation.new_type
        self.schema.put_node_type(current_type)

    async def execute_rename_relationship_type(
        self, operation: RenameRelationshipType
    ) -> None:
        current_type = self.schema.get_relationship_type_by_name(operation.old_type)
        self.schema.drop_relationship_type_by_name(operation.old_type)
        current_type.name = operation.new_type
        self.schema.put_relationship_type(current_type)

    async def execute_add_additional_node_property_index(
        self, operation: AddAdditionalNodePropertyIndex
    ) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.add_index(operation.field_name)

    async def execute_drop_additional_node_property_index(
        self, operation: DropAdditionalNodePropertyIndex
    ) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.drop_index(operation.field_name)

    async def execute_add_additional_relationship_property_index(
        self, operation: AddAdditionalRelationshipPropertyIndex
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.add_index(operation.field_name)

    async def execute_drop_additional_relationship_property_index(
        self, operation: DropAdditionalRelationshipPropertyIndex
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.drop_index(operation.field_name)

    async def execute_add_node_property(self, operation: AddNodeProperty) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.add_property(operation.property_name)

    async def execute_add_relationship_property(
        self, operation: AddRelationshipProperty
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.add_property(operation.property_name)

    async def execute_drop_node_property(self, operation: DropNodeProperty) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.drop_property(operation.property_name)

    async def execute_drop_relationship_property(
        self, operation: DropRelationshipProperty
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.drop_property(operation.property_name)

    async def execute_node_key_extended(self, operation: NodeKeyExtended) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.add_key(operation.added_key_property)

    async def execute_relationship_key_extended(
        self, operation: RelationshipKeyExtended
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.add_key(operation.added_key_property)

    async def execute_node_key_part_renamed(
        self, operation: NodeKeyPartRenamed
    ) -> None:
        type_def = self.schema.get_node_type_by_name(operation.node_type)
        type_def.rename_key(operation.old_key_part_name, operation.new_key_part_name)

    async def execute_relationship_key_part_renamed(
        self, operation: RelationshipKeyPartRenamed
    ) -> None:
        type_def = self.schema.get_relationship_type_by_name(
            operation.relationship_type
        )
        type_def.rename_key(operation.old_key_part_name, operation.new_key_part_name)



from typing import Any, Dict, Optional

from ...pipeline.value_providers import ProviderContext, StaticValueOrValueProvider
from ...schema import GraphObjectSchema, SchemaExpansionCoordinator
from .interpretation import Interpretation
from .property_mapping import PropertyMapping
from .source_node_interpretation import SourceNodeInterpretation


class PropertiesInterpretation(Interpretation, alias="properties"):
    """Stores additional properties onto the source node."""

    __slots__ = ("properties", "norm_args")

    def __init__(
        self,
        properties: StaticValueOrValueProvider,
        normalization: Optional[Dict[str, Any]] = None,
    ):
        self.properties = PropertyMapping.from_file_data(properties)
        self.norm_args = normalization or {}

    def interpret(self, context: ProviderContext):
        source = context.desired_ingest.source
        self.properties.apply_to(context, source.properties, self.norm_args)

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        coordinator.on_node_schema(
            self.expand_source_node_schema,
            alias=SourceNodeInterpretation.SOURCE_NODE_TYPE_ALIAS,
            property_list=list(self.properties),
        )

    def expand_source_node_schema(self, schema: GraphObjectSchema):
        schema.add_properties(self.properties)



import asyncio
from dataclasses import asdict, dataclass, field
from logging import getLogger
from typing import TYPE_CHECKING, List, Tuple

from .creation_rules import NodeCreationRule, RelationshipCreationRule
from .graph_objects import Node, PropertySet, Relationship, RelationshipWithNodes
from .ingestion_hooks import IngestionHook, IngestionHookRunRequest

if TYPE_CHECKING:
    from ..databases.ingest_strategy import IngestionStrategy


LOGGER = getLogger(__name__)


@dataclass(slots=True)
class DesiredIngestion:
    source: Node = field(default_factory=Node)
    relationships: List[RelationshipWithNodes] = field(default_factory=list)
    hook_requests: List[IngestionHookRunRequest] = field(default_factory=list)
    source_node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER

    @property
    def source_node_is_valid(self) -> bool:
        return self.source.is_valid and self.source_node_creation_rule is not None

    async def ingest_source_node(self, strategy: "IngestionStrategy"):
        await strategy.ingest_source_node(self.source, self.source_node_creation_rule)

    async def ingest_relationships(self, strategy: "IngestionStrategy"):
        await asyncio.gather(
            *(
                strategy.ingest_relationship(relationship)
                for relationship in self.relationships
            )
        )

    async def run_ingest_hooks(self, strategy: "IngestionStrategy"):
        await asyncio.gather(
            *(strategy.run_hook(hook_req) for hook_req in self.hook_requests)
        )

    def can_perform_ingest(self):
        # We can do the main part of the ingest if the source node is valid.
        # If it's not valid, it's only an error when there are relationships we are
        # trying to ingest as well.
        if not self.source_node_is_valid:
            if len(self.relationships) > 0:
                LOGGER.warning(
                    "Identity value for source node was null. Skipping Ingest.",
                    extra=asdict(self),
                )
            else:
                LOGGER.debug(
                    "Ingest was not provided a valid source node and no relationships. Only running ingest hooks.",
                    extra=asdict(self),
                )
            return False
        return True

    async def ingest(self, strategy: "IngestionStrategy"):
        if self.can_perform_ingest():
            await self.ingest_source_node(strategy)
            await self.ingest_relationships(strategy)
        await self.run_ingest_hooks(strategy)

    def add_source_node(
        self,
        source_type: str,
        additional_types: Tuple[str],
        creation_rule: NodeCreationRule,
        key_values: PropertySet,
        properties: PropertySet,
    ) -> Node:
        self.source.type = source_type
        self.source.additional_types = additional_types
        self.source_node_creation_rule = creation_rule
        self.source.key_values.merge(key_values)
        self.source.properties.merge(properties)
        # Because relationships can be added before the source node
        self.finalize_relationships()
        return self.source

    def finalize_relationships(self):
        """Finalizes relationships that were added before source node.
        Assumes source node has been added
        """
        for relationship in self.relationships:
            if relationship.outbound:
                relationship.from_node = self.source
                relationship.from_side_node_creation_rule = (
                    self.source_node_creation_rule
                )
            else:
                relationship.to_node = self.source
                relationship.to_side_node_creation_rule = self.source_node_creation_rule

    def add_relationship(
        self,
        related_node: Node,
        relationship: Relationship,
        outbound: bool,
        node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER,
        relationship_creation_rule: RelationshipCreationRule = RelationshipCreationRule.EAGER,
    ):
        if not related_node.is_valid:
            LOGGER.warning(
                "Identity value for related node was null. Skipping.",
                extra=asdict(related_node),
            )
            return

        if outbound:
            from_node, from_match = (self.source, self.source_node_creation_rule)
            to_node, to_match = (related_node, node_creation_rule)
        else:
            from_node, from_match = (related_node, node_creation_rule)
            to_node, to_match = (self.source, self.source_node_creation_rule)

        self.relationships.append(
            RelationshipWithNodes(
                from_node=from_node,
                to_node=to_node,
                outbound=outbound,
                relationship=relationship,
                from_side_node_creation_rule=from_match,
                to_side_node_creation_rule=to_match,
                relationship_creation_rule=relationship_creation_rule,
            )
        )

    def add_ingest_hook(self, hook: IngestionHook, before_ingest=False):
        self.hook_requests.append(IngestionHookRunRequest(hook, before_ingest))



from typing import Any, Dict, Iterable

from ...model import TimeToLiveConfiguration
from ...schema import GraphObjectType
from .iterable import IterableExtractor


class TimeToLiveConfigurationExtractor(IterableExtractor):
    def __init__(
        self,
        graph_object_type: str,
        configurations: Iterable[Dict[str, Any]],
        override_expiry_in_hours: int = None,
    ) -> None:

        def make_config(config: Dict[str, Any]) -> None:
            if override_expiry_in_hours is not None:
                config["expiry_in_hours"] = override_expiry_in_hours
            return TimeToLiveConfiguration(
                **config,
                graph_object_type=GraphObjectType(graph_object_type),
            )

        super().__init__([make_config(config) for config in configurations])



from logging import getLogger
from typing import Any, AsyncGenerator, Iterable

from .extractor import Extractor


class IterableExtractor(Extractor):
    """An extractor that produces records from an iterable."""

    @classmethod
    def range(cls, start=0, stop=100, step=1):
        return cls(iterable=({"index": i} for i in range(start, stop, step)))

    def __init__(self, iterable: Iterable[Any]) -> None:
        self.iterable = iterable
        self.index = 0
        self.logger = getLogger(self.__class__.__name__)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for index, record in enumerate(self.iterable):
            if index < self.index:
                continue
            self.index = index
            yield record

    async def make_checkpoint(self):
        return self.index

    async def resume_from_checkpoint(self, checkpoint):
        if isinstance(checkpoint, int):
            self.index = checkpoint
            self.logger.info(f"Resuming from checkpoint {checkpoint}")



from typing import Iterable, Optional

from yaml import safe_dump

from ...metrics import Metrics
from ...pipeline import PipelineInitializationArguments, PipelineProgressReporter
from ...project import Project, RunRequest
from ...project.pipeline_definition import PipelineDefinition
from ...utils import StringSuggester
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

STATS_TABLE_COLS = ["Statistic", "Value"]

ERROR_NO_PIPELINES_FOUND = "<error>No pipelines with the provided name were found in your project. If you didn't provide a name, you have no pipelines.</error>"
HINT_CHECK_PIPELINE_NAME = "<info>HINT: Check that the pipelines you are trying to run are named correctly and in the registry.</info>"
HINT_USE_NODESTREAM_SHOW = "<info>HINT: You can view your project's pipelines by running 'nodestream show'. </info>"
WARNING_NO_TARGETS_PROVIDED = "<error>No targets provided. Running pipeline without writing to any targets.</error>"


class RunPipeline(Operation):
    def __init__(self, project: Project) -> None:
        self.project = project

    def get_pipeline_to_run_or_suggest_to_user(
        self, command: NodestreamCommand, pipeline_name: str
    ) -> Optional[PipelineDefinition]:
        if pipeline_definition := self.project.get_pipeline_by_name(pipeline_name):
            return pipeline_definition

        all_names = {ppl.name for ppl in self.project.get_all_pipelines()}
        suggester = StringSuggester(all_names)
        suggested_name = suggester.suggest_closest(pipeline_name)
        command.line(
            f"<error>No pipeline with the name '{pipeline_name}' found in your project</error>"
        )
        command.line(f"<info>HINT: Did you mean '{suggested_name}'?</info>")

    def get_pipelines_to_run(
        self, command: NodestreamCommand
    ) -> Iterable[PipelineDefinition]:
        supplied_commands = command.argument("pipelines")
        if supplied_commands:
            return [
                ppl
                for name in supplied_commands
                if (ppl := self.get_pipeline_to_run_or_suggest_to_user(command, name))
            ]
        return self.project.get_all_pipelines()

    async def perform(self, command: NodestreamCommand):
        pipelines_ran = 0
        for pipeline in self.get_pipelines_to_run(command):
            request = self.make_run_request(command, pipeline)
            pipelines_ran += await self.project.run(request)

        if pipelines_ran == 0:
            command.line(ERROR_NO_PIPELINES_FOUND)
            command.line(HINT_CHECK_PIPELINE_NAME)
            command.line(HINT_USE_NODESTREAM_SHOW)

    def get_writer_steps_for_specified_targets(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ):
        targets_names = self.combine_targets_from_command_and_pipeline(
            command, pipeline
        )
        for target_name in targets_names:
            target = self.project.get_target_by_name(target_name)
            if target:
                yield target.make_writer()
            else:
                command.line(
                    f"<error>Target '{target_name}' not found in project. Ignoring.</error>"
                )

    def combine_targets_from_command_and_pipeline(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ):
        from_cli = set(command.option("target") or {})
        from_pipeline = set(pipeline.configuration.effective_targets or {})
        effective_target_names = from_cli.union(from_pipeline)
        if not effective_target_names:
            command.line(WARNING_NO_TARGETS_PROVIDED)
        return effective_target_names

    def make_run_request(
        self, command: NodestreamCommand, pipeline: PipelineDefinition
    ) -> RunRequest:
        def print_effective_config(config):
            if command.is_very_verbose:
                command.line("<info>Effective configuration:</info>")
                command.line(f"<info>{safe_dump(config)}</info>")

        storage_name = command.option("storage-backend")
        object_store = self.project.get_object_storage_by_name(storage_name)

        return RunRequest(
            pipeline_name=pipeline.name,
            initialization_arguments=PipelineInitializationArguments(
                annotations=command.option("annotations"),
                step_outbox_size=int(command.option("step-outbox-size")),
                on_effective_configuration_resolved=print_effective_config,
                extra_steps=list(
                    self.get_writer_steps_for_specified_targets(command, pipeline)
                ),
                object_store=object_store,
            ),
            progress_reporter=self.create_progress_reporter(command, pipeline.name),
        )

    def get_progress_indicator(
        self, command: NodestreamCommand, pipeline_name: str
    ) -> "ProgressIndicator":
        if command.has_json_logging_set:
            return ProgressIndicator(command, pipeline_name)

        return SpinnerProgressIndicator(command, pipeline_name)

    def create_progress_reporter(
        self, command: NodestreamCommand, pipeline_name: str
    ) -> PipelineProgressReporter:
        indicator = self.get_progress_indicator(command, pipeline_name)
        return PipelineProgressReporter(
            reporting_frequency=int(command.option("reporting-frequency")),
            callback=indicator.progress_callback,
            on_start_callback=indicator.on_start,
            on_finish_callback=indicator.on_finish,
            on_fatal_error_callback=indicator.on_fatal_error,
        )


class ProgressIndicator:
    def __init__(self, command: NodestreamCommand, pipeline_name: str) -> None:
        self.command = command
        self.pipeline_name = pipeline_name

    def on_start(self):
        pass

    def progress_callback(self, _, __):
        pass

    def on_finish(self, metrics: Metrics):
        pass

    def on_fatal_error(self, exception: Exception):
        self.command.line(
            "<error>Encountered a fatal error while running pipeline</error>"
        )
        self.command.line(f"<error>{exception}</error>")


class SpinnerProgressIndicator(ProgressIndicator):
    def __init__(self, command: NodestreamCommand, pipeline_name: str) -> None:
        super().__init__(command, pipeline_name)
        self.exception = None

    def on_start(self):
        self.progress = self.command.progress_indicator()
        self.progress.start(f"Running pipeline: '{self.pipeline_name}'")

    def progress_callback(self, index, _):
        self.progress.set_message(
            f"Currently processing record at index: <info>{index}</info>"
        )

    def on_finish(self, metrics: Metrics):
        self.progress.finish(f"Finished running pipeline: '{self.pipeline_name}'")

        if self.exception:
            raise self.exception

    def on_fatal_error(self, exception: Exception):
        self.progress.set_message(
            "<error>Encountered a fatal error while running pipeline</error>"
        )
        self.exception = exception



import asyncio
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import Any, AsyncGenerator, Dict, Optional

from ..class_loader import ClassLoader
from ..flush import Flush
from ..step import Step, StepContext
from ..value_providers import ProviderContext, StaticValueOrValueProvider, ValueProvider


class Transformer(Step):
    """A `Transformer` takes a given record and mutates into a new record.

    `Transformer` steps generally make up the middle of an ETL pipeline and
    are responsible for reshaping an object so its more ingestible by the
    downstream sink.
    """

    async def process_record(self, record, _) -> AsyncGenerator[object, None]:
        if record is Flush:
            yield record
        else:
            val_or_gen = self.transform_record(record)
            if isinstance(val_or_gen, AsyncGenerator):
                async for result in val_or_gen:
                    yield result
            else:
                yield await val_or_gen

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError


class ConcurrentTransformer(Transformer, ABC):
    """A `ConcurrentTransformer` can process multiple records concurrently.

    `ConcurrentTransformer`s are useful when the transformation work is IO
    bound and can be parallelized across multiple threads.
    `ConcurrentTransformer`s are not useful when the transformation work is
    CPU bound as the GIL will prevent useful parallelization. Additionally,
    `ConcurrentTransformer`s do not guarantee the order of the output stream
    will match the input stream.
    """

    def __init__(
        self,
        thread_pool_size: Optional[int] = None,
        maximum_pending_tasks: int = 1000,
    ) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.maximum_pending_tasks = maximum_pending_tasks
        self.thread_pool = ThreadPoolExecutor(
            max_workers=thread_pool_size,
            thread_name_prefix=self.__class__.__name__,
            initializer=self.on_worker_thread_start,
        )
        self.pending_tasks = []

    def on_worker_thread_start(self):
        pass  # do nothing

    async def drain_completed_tasks(self):
        tasks_drained = 0
        remaining_pending_tasks = []
        for task in self.pending_tasks:
            if task.done():
                yield task.result()
                tasks_drained += 1
            else:
                remaining_pending_tasks.append(task)

        await self.yield_processor()

        self.pending_tasks[:] = remaining_pending_tasks
        if tasks_drained:
            self.logger.debug(
                "Drained %s completed tasks, %s pending tasks remain",
                tasks_drained,
                len(self.pending_tasks),
            )

    async def process_record(self, record: Any, _) -> Any:
        if record is Flush:
            async for outstanding in self.drain_completed_tasks():
                yield outstanding
            yield record
        else:
            # Submit the work to the thread pool (only if we have capacity)
            # If we don't have capacity, yield completed tasks.
            # Once there is nothing remaining to yield from these tasks,
            # yield the processor until capacity is available.
            submitted = False
            while not submitted:
                if len(self.pending_tasks) < self.maximum_pending_tasks:
                    task = self.thread_pool.submit(self.do_work_on_record, record)
                    self.pending_tasks.append(task)
                    submitted = True
                async for result in self.drain_completed_tasks():
                    yield result

    def do_work_on_record(self, record: Any) -> Any:
        return asyncio.run(self.transform_record(record))

    async def yield_processor(self):
        await asyncio.sleep(0)

    async def emit_outstanding_records(self, context: StepContext):
        while self.pending_tasks:
            async for result in self.drain_completed_tasks():
                yield result

    async def finish(self, _: StepContext):
        await asyncio.get_event_loop().run_in_executor(
            None, self.thread_pool.shutdown, True
        )


class PassTransformer(Transformer):
    async def transform_record(self, record: Any) -> AsyncGenerator:
        yield record


class SwitchTransformer(Transformer):
    __slots__ = (
        "switch_on",
        "transformers",
        "default",
        "normalization",
    )

    @classmethod
    def from_file_data(cls, switch_on, cases, default=None, normalization=None):
        loader = ClassLoader.instance(Step)
        transformers = {
            field_value: loader.load_class(**transformer)
            for field_value, transformer in cases.items()
        }
        default = loader.load_class(**default) if default else None
        return cls(switch_on, transformers, default, normalization)

    def __init__(
        self,
        switch_on: StaticValueOrValueProvider,
        transformers: Dict[str, Transformer],
        default: Transformer = None,
        normalization: Dict[str, Any] = None,
    ):
        self.switch_on = ValueProvider.guarantee_value_provider(switch_on)
        self.transformers = transformers
        self.default = default if default else PassTransformer()
        self.normalization = normalization or {}

    async def transform_record(self, record: Any):
        context = ProviderContext.fresh(record)
        key = self.switch_on.normalize_single_value(context, self.normalization)
        transformer = self.transformers.get(key, self.default)
        # If we don't have a transformer to handle the record we just yield
        # the record with PassTransformer
        async for result in transformer.transform_record(record):
            yield result



from dataclasses import dataclass, field
from difflib import SequenceMatcher
from typing import Dict, Generator, Generic, Iterable, List, TypeVar


class StringSuggester:
    """A utility for suggesting alternative strings based on similarity to a given string."""

    def __init__(self, all_strings: Iterable[str]):
        self.all_strings = list(all_strings)

    def suggest_closest(self, string: str) -> str:
        """Suggest the string that is most similar to the given string.

        Args:
            string: The string to compare to the list of strings.

        Returns:
            The string from the list of strings that is most similar to the given string.
        """
        return max(
            self.all_strings,
            key=lambda s: SequenceMatcher(None, string, s).ratio(),
        )


T = TypeVar("T")

"""
    LayeredList Datatype:
        This datatype mimicks the list datatype in python to the requestor, but involves different 
        layers of data that is aggregated together on reads, and only reveals the top layer on writes. 
"""


@dataclass(slots=True, frozen=True)
class LayeredList(Generic[T]):
    LATEST_CONTEXT_LEVEL = -1

    _data: List[List[T]] = field(default_factory=lambda: [[]])

    def __setitem__(self, index: int, item: T) -> None:
        self._data[self.LATEST_CONTEXT_LEVEL][index] = item

    def __getitem__(self, index: int) -> T:
        return self.effective_items[index]

    def __contains__(self, value: T) -> bool:
        return value in self.effective_items

    def pop(self, index: int = -1) -> T:
        return self._data[self.LATEST_CONTEXT_LEVEL].pop(index)

    def increment_context_level(self) -> None:
        self._data.append([])

    def decrement_context_level(self) -> None:
        self._data.pop()

    def append(self, data: T) -> None:
        self._data[self.LATEST_CONTEXT_LEVEL].append(data)

    def __iter__(self) -> Generator[T, None, None]:
        yield from self.effective_items

    def __del__(self) -> None:
        self.decrement_context_level()

    @property
    def effective_items(self) -> List[T]:
        return_value = []
        for list_obj in self._data:
            return_value += list_obj
        return return_value


K = TypeVar("K")
V = TypeVar("V")

"""
    LayeredDict Datatype:
        This datatype mimicks the dictionary datatype in python to the requestor, but involves different 
        layers of data that is aggregated together on reads, and only reveals the top layer on writes. 
"""


@dataclass(slots=True, frozen=True)
class LayeredDict(Generic[K, V]):
    LATEST_CONTEXT_LEVEL = -1

    _data: List[Dict[K, V]] = field(default_factory=lambda: [{}])

    def __setitem__(self, key: K, item: V):
        self._data[self.LATEST_CONTEXT_LEVEL][key] = item

    def __getitem__(self, key: K) -> V:
        return self.effective_items[key]

    def __contains__(self, key: K) -> bool:
        return key in self.effective_items

    def increment_context_level(self) -> None:
        self._data.append({})

    def decrement_context_level(self) -> None:
        self._data.pop()

    def pop(self, key: K, default: V) -> V:
        return self._data[self.LATEST_CONTEXT_LEVEL].pop(key, default)

    def get(self, key: K, default: V) -> V:
        return self.effective_items.get(key, default)

    def items(self) -> Generator[tuple[K, V], None, None]:
        yield from ([k, v] for k, v in self.effective_items.items())

    def values(self) -> Generator[V, None, None]:
        yield from (v for v in self.effective_items.values())

    def keys(self) -> Generator[V, None, None]:
        yield from (k for k in self.effective_items.keys())

    def __del__(self) -> None:
        self.decrement_context_level()

    @property
    def effective_items(self):
        return_value = {}
        for dict_obj in self._data:
            for key, value in dict_obj.items():
                return_value[key] = value
        return return_value



from typing import AsyncGenerator, Optional

from ..metrics import Metric, Metrics
from .object_storage import ObjectStore
from .progress_reporter import PipelineProgressReporter


class StepContext:
    """`StepContext` is a context object that is passed to steps in a pipeline.

    The `StepContext` provides a way for steps to interact with the pipeline
    and report and perist information about the state of the pipeline.
    """

    __slots__ = ("reporter", "index", "name", "object_store")

    def __init__(
        self,
        name: str,
        index: int,
        reporter: PipelineProgressReporter,
        object_store: ObjectStore,
    ) -> None:
        self.name = name
        self.reporter = reporter
        self.index = index
        self.object_store = object_store

    @property
    def pipeline_encountered_fatal_error(self) -> bool:
        """Whether the pipeline has encountered a fatal error."""
        return self.reporter.encountered_fatal_error

    def report_error(
        self,
        message: str,
        exception: Optional[Exception] = None,
        fatal: bool = False,
        **extras,
    ):
        """Report an error.

        This method is used to report an error that occurred during the
        processing of a step. It can be used to log the error and take
        appropriate action based on the error.

        Args:
            message: The error message.
            exception: The exception that caused the error.
            fatal: Whether the error is fatal.
        """
        self.reporter.logger.error(
            message,
            extra=dict(index=self.index, fatal=fatal, step_name=self.name, **extras),
            exc_info=exception,
            stack_info=True,
        )
        if fatal:
            self.reporter.on_fatal_error(exception)
        else:
            Metrics.get().increment(Metric.NON_FATAL_ERRORS)

    def debug(self, message: str, **extras):
        """Log a debug message.

        This method is used to log a debug message. It can be used to log
        information about the state of the pipeline and the steps in the
        pipeline.
        """
        self.reporter.logger.debug(
            message, extra=dict(index=self.index, step_name=self.name, **extras)
        )

    def info(self, message: str, **extras):
        """Log an info message.

        This method is used to log an info message. It can be used to log
        information about the state of the pipeline and the steps in the
        pipeline.
        """
        self.reporter.logger.info(
            message, extra=dict(index=self.index, step_name=self.name, **extras)
        )

    def warning(self, message: str, **extras):
        """Log a warning message.

        This method is used to log a warning message. It can be used to log
        information about the state of the pipeline and the steps in the
        pipeline.
        """
        self.reporter.logger.warning(
            message, extra=dict(index=self.index, step_name=self.name, **extras)
        )


class Step:
    """A `Step` is a unit of work that can be executed in a pipeline.

    Steps are the building blocks of a pipeline. They are responsible for
    processing records and emitting new records. Steps can be chained together
    to form a pipeline.

    Steps are asynchronous and can be started, stopped, and run in an
    asynchronous context. They can process records and emit new records.
    """

    async def start(self, context: StepContext):
        """Start the step.

        This method is called when the step is started. It is responsible for
        setting up the step and preparing it for processing records. This
        method is called once before any records are processed.
        """
        pass

    async def process_record(
        self, record, context: StepContext
    ) -> AsyncGenerator[object, None]:
        """Process a record.

        This method is called for each record that is passed to the step. It
        should process the record and emit new records.
        """
        yield record

    async def emit_outstanding_records(self, context: StepContext):
        """Emit any outstanding records.

        This method is called after all records have been processed. It is
        responsible for emitting any outstanding records that were not emitted
        during the processing of records.
        """
        for record in ():
            yield record  # pragma: no cover

    async def finish(self, context: StepContext):
        """Finish the step.

        This method is called when the step is finished. It is responsible for
        cleaning up the step and releasing any resources that were acquired
        during the processing of records. This method is called once after all
        records have been processed.
        """
        pass


class PassStep(Step):
    """A `PassStep` passes records through."""

    def __init__(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)



from typing import Any

from .normalizer import Normalizer


class UppercaseStrings(Normalizer, alias="uppercase_strings"):
    def normalize_value(self, value: Any) -> Any:
        return value.upper() if isinstance(value, str) else value



from abc import ABC, abstractmethod
from typing import Any, Iterable, Type

import jmespath
from jmespath.parser import ParsedResult
from yaml import SafeDumper, SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider, ValueProviderException

# `QueryStrategy` is here to provide the seam for different optimizations
# for executing jmespath queries. We can either execute a "fully fledged"
# jmespath query or we can implement some simple access patterns that
# are faster to execute. For example, if the expression is a simple key
# lookup, we can just use the key directly instead of compiling the
# jmespath expression and then executing it with all the weight and
# overhead that comes with it.


class QueryStrategy(ABC):
    @classmethod
    def from_string_expression(cls, expression: str):
        if expression.isalpha():
            return KeyLookup(expression)

        compiled_query = jmespath.compile(expression)
        return ExecuteJmespath(compiled_query)

    @abstractmethod
    def search(self, context: ProviderContext):
        pass


class ExecuteJmespath(QueryStrategy):
    def __init__(self, compiled_query: ParsedResult) -> None:
        self.compiled_query = compiled_query

    def search(self, context: ProviderContext):
        return self.compiled_query.search(context.document)

    def __str__(self) -> str:
        return str(self.compiled_query.expression)


class KeyLookup(QueryStrategy):
    def __init__(self, key: str) -> None:
        self.key = key

    def search(self, context: ProviderContext):
        return context.document.get(self.key, None)

    def __str__(self) -> str:
        return self.key


class JmespathValueProvider(ValueProvider):
    """A `ValueProvider` that uses JMESPath to extract values from a document."""

    __slots__ = ("compiled_query",)

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!jmespath",
            lambda loader, node: cls.from_string_expression(
                loader.construct_scalar(node)
            ),
        )

    @classmethod
    def from_string_expression(cls, expression: str):
        return cls(QueryStrategy.from_string_expression(expression))

    def __init__(self, strategy: QueryStrategy) -> None:
        self.strategy = strategy

    def search(self, context: ProviderContext):
        raw_search = self.strategy.search(context)
        if raw_search is None:
            return
        if isinstance(raw_search, list):
            yield from raw_search
        else:
            yield raw_search

    def single_value(self, context: ProviderContext) -> Any:
        try:
            return next(self.search(context), None)
        except Exception as e:
            raise ValueProviderException(str(context.document), self) from e

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        try:
            yield from self.search(context)
        except Exception as e:
            raise ValueProviderException(str(context.document), self) from e

    def __str__(self) -> str:
        return f"JmespathValueProvider: { {'expression': str(self.strategy)} }"


SafeDumper.add_representer(
    JmespathValueProvider,
    lambda dumper, jmespath: dumper.represent_scalar(
        "!jmespath", str(jmespath.strategy)
    ),
)



from pathlib import Path
from typing import AsyncIterable, Optional, Tuple

from ..state import Schema
from .auto_change_detector import AutoChangeDetector, MigratorInput
from .auto_migration_maker import AutoMigrationMaker
from .migrations import Migration, MigrationGraph
from .migrator import Migrator
from .state_providers import MigrationGraphStateProvider, StaticStateProvider


class ProjectMigrations:
    """The migrations for a project.

    This class is used to be the interface between the project and the schema
    migrations system. It is used to execute migrations and generate migrations
    from changes in the project schema.
    """

    def __init__(self, graph: MigrationGraph, source_directory: Path) -> None:
        self.graph = graph
        self.source_directory = source_directory

    @classmethod
    def from_directory(cls, directory: Path) -> "ProjectMigrations":
        """Load migrations from a directory.

        Args:
            directory: The directory to load migrations from.

        Returns:
            The loaded migrations.
        """
        graph = MigrationGraph.from_directory(directory)
        return cls(graph, directory)

    async def determine_pending(
        self, migrator: Migrator
    ) -> AsyncIterable[Tuple[Migration, bool]]:
        """Determine pending migrations.

        This method determines the pending migrations for a migrator. A pending
        migration is a migration that has not been executed yet.

        Args:
            migrator: The migrator to use to determine pending migrations.

        Yields:
            Each migration and a boolean indicating if it is pending.
        """
        completed_migrations = await migrator.get_completed_migrations(self.graph)
        for migration in self.graph.topological_order():
            yield migration, migration not in completed_migrations

    async def execute_pending(self, migrator: Migrator) -> AsyncIterable[Migration]:
        """Execute pending migrations.

        This method executes all pending (non-completed) migrations.

        Args:
            migrator: The migrator to use to execute migrations.

        Yields:
            The names of the migrations that were executed.
        """
        async for migration, pending in self.determine_pending(migrator):
            if pending:
                await migrator.execute_migration(migration)
                yield migration

    async def detect_changes(
        self, input: MigratorInput, current_state: Schema
    ) -> Optional[Migration]:
        """Detect changes in the project schema.

        This method detects changes in the project schema compared to the
        current migration state. If there are no changes, then this method
        returns None.

        Args:
            input: The input interface for the migrator to ask questions.
            project_schema: The schema of the project in its current state.

        Returns:
            The generated migration, or None if there were no changes.
        """
        detector = AutoChangeDetector(
            input,
            MigrationGraphStateProvider(self.graph),
            StaticStateProvider(current_state),
        )
        maker = AutoMigrationMaker(self.graph, detector)
        return await maker.make_migration()

    async def create_migration_from_changes(
        self, input: MigratorInput, current_state: Schema
    ) -> Optional[Tuple[Migration, Path]]:
        """Create a migration from changes.

        Calls `detect_changes` and if there are changes, then it creates a
        migrtation file in the project and returns the migration and the path
        to the file. If there are no changes, then this method returns None.

        Args:
            input: The input interface for the migrator to ask questions.
            project_schema: The schema of the project in its current state.

        Returns:
            The generated migration and path to it, or None if there were no changes.
        """
        if (migration := await self.detect_changes(input, current_state)) is None:
            return None
        path = migration.write_to_file_with_default_name(self.source_directory)
        return migration, path

    def create_squash_between(
        self, from_migration: Migration, to_migration: Optional[Migration] = None
    ) -> Tuple[Migration, Path]:
        """Create a squashed migration between two migrations.

        Args:
            from_migration: The migration to squash from.
            to_migration: The migration to squash to.

        Returns:
            The squashed migration and the path to the file.
        """
        if to_migration is None:
            name = "squash_from_{}".format(from_migration.name)
        else:
            name = "squash_from_{}_to_{}".format(from_migration.name, to_migration.name)
        squashed = self.graph.squash_between(name, from_migration, to_migration)
        path = squashed.write_to_file_with_default_name(self.source_directory)
        return squashed, path



from typing import AsyncGenerator, Iterable

from ..model import IngestionHook, Node, RelationshipWithNodes, TimeToLiveConfiguration
from ..schema.migrations import Migrator
from ..schema.migrations.operations import Operation
from .copy import TypeRetriever
from .database_connector import DatabaseConnector
from .query_executor import (
    OperationOnNodeIdentity,
    OperationOnRelationshipIdentity,
    QueryExecutor,
)


async def empty_async_generator():
    for i in []:
        yield i  # pragma: no cover


class NullMigrator(Migrator):
    async def execute_operation(self, _: Operation):
        pass


class NullQueryExecutor(QueryExecutor):
    async def upsert_nodes_in_bulk_with_same_operation(
        self, _: OperationOnNodeIdentity, __: Iterable[Node]
    ):
        pass

    async def upsert_relationships_in_bulk_of_same_operation(
        self,
        _: OperationOnRelationshipIdentity,
        __: Iterable[RelationshipWithNodes],
    ):
        pass

    async def perform_ttl_op(self, _: TimeToLiveConfiguration):
        pass

    async def execute_hook(self, _: IngestionHook):
        pass

    async def finish(self):
        """Close connector by calling finish method from Step"""
        pass


class NullRetriver(TypeRetriever):
    def get_nodes_of_type(self, _: str) -> AsyncGenerator[Node, None]:
        return empty_async_generator()

    def get_relationships_of_type(
        self, _: str
    ) -> AsyncGenerator[RelationshipWithNodes, None]:
        return empty_async_generator()


class NullConnector(DatabaseConnector, alias="null"):
    def __init__(self, **_) -> None:
        pass

    def make_migrator(self) -> TypeRetriever:
        return NullMigrator()

    def make_query_executor(self) -> QueryExecutor:
        return NullQueryExecutor()

    def make_type_retriever(self) -> TypeRetriever:
        return NullRetriver()



from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Dict, Iterable, Optional, Set, Tuple

from nodestream.utils import LayeredDict, LayeredList

from ..file_io import LoadsFromYaml, LoadsFromYamlFile, SavesToYaml, SavesToYamlFile


class GraphObjectType(str, Enum):
    """Represents the type of a graph object."""

    NODE = "NODE"
    RELATIONSHIP = "RELATIONSHIP"


class Cardinality(str, Enum):
    """Represents the cardinality of an adjacency."""

    SINGLE = "SINGLE"
    MANY = "MANY"


class PropertyType(str, Enum):
    """Represents the type of a property."""

    STRING = "STRING"
    BOOLEAN = "BOOLEAN"
    DATETIME = "DATETIME"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"

    # This is definitely not ideal to describe types this way but it works for now.
    LIST_OF_STRINGS = "LIST_OF_STRINGS"
    LIST_OF_BOOLEANS = "LIST_OF_BOOLEANS"
    LIST_OF_DATETIMES = "LIST_OF_DATETIMES"
    LIST_OF_INTEGERS = "LIST_OF_INTEGERS"
    LIST_OF_FLOATS = "LIST_OF_FLOATS"


@dataclass(slots=True)
class PropertyMetadata(LoadsFromYaml, SavesToYaml):
    """Metadata for a property.

    Attributes:
        property_type: The type of the property.
    """

    type: PropertyType = PropertyType.STRING
    is_key: bool = False
    is_indexed: bool = False

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                "type": str,
                Optional("is_key"): bool,
                Optional("is_indexed"): bool,
            }
        )

    @classmethod
    def from_file_data(cls, yaml_data):
        return cls(
            type=PropertyType(yaml_data["type"]),
            is_key=yaml_data.get("is_key", False),
            is_indexed=yaml_data.get("is_indexed", False),
        )

    def to_file_data(self):
        return {
            "type": self.type.value,
            "is_key": self.is_key,
            "is_indexed": self.is_indexed,
        }

    def merge(self, other: "PropertyMetadata"):
        """Merge another property metadata into this property metadata.

        Args:
            other: The other property metadata.
        """
        self.type = other.type
        self.is_key |= other.is_key
        self.is_indexed |= other.is_indexed


@dataclass(slots=True)
class GraphObjectSchema(LoadsFromYaml, SavesToYaml):
    """A graph object type in a schema."""

    name: str
    properties: Dict[str, PropertyMetadata] = field(default_factory=dict)

    def __str__(self) -> str:
        return f"{self.name}:\n" + "\n".join(
            f"  {property_name}: {property_metadata.type.value}"
            for property_name, property_metadata in self.properties.items()
        )

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema(
            {
                "name": str,
                "properties": {str: PropertyMetadata.describe_yaml_schema()},
            }
        )

    @classmethod
    def from_file_data(cls, yaml_data):
        return cls(
            name=yaml_data["name"],
            properties={
                key: PropertyMetadata.from_file_data(value)
                for key, value in yaml_data["properties"].items()
            },
        )

    def to_file_data(self):
        return {
            "name": self.name,
            "properties": {
                key: value.to_file_data() for key, value in self.properties.items()
            },
        }

    @property
    def keys(self) -> Set[str]:
        return {
            property_name
            for property_name, metadata in self.properties.items()
            if metadata.is_key
        }

    @property
    def non_key_properties(self) -> Set[str]:
        return set(self.properties) - self.keys

    @property
    def indexed_properties(self) -> Set[str]:
        return {
            property_name
            for property_name, metadata in self.properties.items()
            if metadata.is_indexed
        }

    def rename_property(self, old_property_name: str, new_property_name: str):
        """Rename a property.

        Args:
            old_property_name: The old property name.
            new_property_name: The new property name.
        """
        if old_property_name not in self.properties:
            raise ValueError(
                f"Property {old_property_name} does not exist on node type {self.name}."
            )

        self.properties[new_property_name] = self.properties.pop(old_property_name)

    def rename_key(self, old_key: str, new_key: str):
        """Rename a key.

        Args:
            old_key: The old key.
            new_key: The new key.
        """
        if old_key not in self.keys:
            raise ValueError(f"Key {old_key} does not exist on node type {self.name}.")
        self.rename_property(old_key, new_key)

    def add_index(self, property: str, auto_add_property: bool = True):
        """Add an index.

        Args:
            index: The index to add.
        """
        if property not in self.properties and auto_add_property:
            self.add_property(property)
        self.properties[property].is_indexed = True

    def drop_index(self, property: str):
        """Drop an index.

        Args:
            index: The index to drop.
        """
        self.properties[property].is_indexed = False

    def has_matching_keys(
        self, other: "GraphObjectSchema", allow_undefined: bool = False
    ) -> bool:
        """Check if two node types have matching keys.

        Args:
            other: The other node type.
            allow_undefined: Whether or not to allow undefined keys. If this is
                True, then the node types will be considered to have matching
                keys if one of the node types has no keys.

        Returns:
            True if the node types have matching keys.
        """
        us = self.keys
        them = other.keys
        return us == them or (allow_undefined and len(us) == 0 or len(them) == 0)

    def has_matching_properties(self, other: "GraphObjectSchema") -> bool:
        """Check if two node types have matching properties.

        Args:
            other: The other node type.

        Returns:
            True if the node types have matching properties.
        """
        return self.properties == other.properties

    def add_property(
        self, property_name: str, metadata: Optional[PropertyMetadata] = None
    ):
        """Add a property.

        If the property already exists, then the property will be replaced.
        If no metadata is provided, then the property will be added with the
        default metadata (string type, not key, not indexed).

        Args:
            property_name: The property name.
        """
        metadata = metadata or PropertyMetadata()
        if property_name in self.properties:
            self.properties[property_name].merge(metadata)
        else:
            self.properties[property_name] = metadata

    def drop_property(self, property_name: str):
        """Drop a property.

        Args:
            property_name: The property name.
        """
        del self.properties[property_name]

    def add_key(self, key: str):
        """Add a key.

        Args:
            key: The key.
        """
        metadata = PropertyMetadata(PropertyType.STRING, is_key=True)
        self.add_property(key, metadata)

    def add_keys(self, keys: Iterable[str]):
        """Add keys.

        Args:
            keys: The keys.
        """
        if self.keys and self.keys != set(keys):
            raise ValueError(
                f"Cannot add keys {keys} to node type {self.name} because it already has keys {self.keys}."
            )

        for key in keys:
            self.add_key(key)

    def add_properties(self, properties: Iterable[str]):
        """Add properties.

        Args:
            properties: The properties.
        """
        for property in properties:
            self.add_property(property)

    def add_indexes(self, indexes: Iterable[str]):
        """Add indexes to the properties of this node type.

        Args:
            indexes: The properties to index.
        """
        for property in indexes:
            self.add_index(property)

    def add_indexed_timestamp(self):
        """Add a timestamp index."""
        property_name = "last_ingested_at"
        metadata = PropertyMetadata(PropertyType.DATETIME)
        self.add_property(property_name, metadata)
        self.add_index(property_name)

    def merge(self, other: "GraphObjectSchema"):
        """Merge another node type into this node type.

        Args:
            other: The other node type.
        """
        if self.name != other.name:
            raise ValueError(
                f"Cannot merge node type {other.name} into node type {self.name}."
            )

        if not self.has_matching_keys(other, allow_undefined=True):
            raise ValueError(
                f"Cannot merge node type {other.name} into node type {self.name} because the keys do not match."
            )

        for property_name, metadata in other.properties.items():
            if property_name not in self.properties:
                self.properties[property_name] = metadata
            else:
                self.properties[property_name].merge(metadata)


@dataclass(slots=True, frozen=True)
class Adjacency:
    """An adjacency between two node types."""

    from_node_type: str
    to_node_type: str
    relationship_type: str

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema(
            {"from_node_type": str, "to_node_type": str, "relationship_type": str}
        )

    @classmethod
    def from_file_data(cls, yaml_data):
        return cls(
            from_node_type=yaml_data["from_node_type"],
            to_node_type=yaml_data["to_node_type"],
            relationship_type=yaml_data["relationship_type"],
        )

    def to_file_data(self):
        return {
            "from_node_type": self.from_node_type,
            "to_node_type": self.to_node_type,
            "relationship_type": self.relationship_type,
        }


@dataclass(slots=True, frozen=True)
class AdjacencyCardinality:
    """An adjacency cardinality."""

    from_side_cardinality: Cardinality = Cardinality.SINGLE
    to_side_cardinality: Cardinality = Cardinality.SINGLE

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema(
            {
                "from_side_cardinality": str,
                "to_side_cardinality": str,
            }
        )

    @classmethod
    def from_file_data(cls, yaml_data):
        return cls(
            from_side_cardinality=Cardinality(yaml_data["from_side_cardinality"]),
            to_side_cardinality=Cardinality(yaml_data["to_side_cardinality"]),
        )

    def to_file_data(self):
        return {
            "from_side_cardinality": self.from_side_cardinality,
            "to_side_cardinality": self.to_side_cardinality,
        }


@dataclass(slots=True, frozen=True)
class Schema(SavesToYamlFile, LoadsFromYamlFile):
    """A schema for a database."""

    type_schemas: Dict[Tuple[GraphObjectType, str], GraphObjectSchema] = field(
        default_factory=dict
    )

    cardinalities: Dict[Adjacency, AdjacencyCardinality] = field(default_factory=dict)

    def __str__(self):
        return "\n".join(
            f"[{type.value}] {str(s)}" for (type, _), s in self.type_schemas.items()
        )

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                Optional("nodes"): [GraphObjectSchema.describe_yaml_schema()],
                Optional("relationships"): [GraphObjectSchema.describe_yaml_schema()],
                Optional("cardinalities"): [
                    {
                        "adjacency": Adjacency.describe_yaml_schema(),
                        "cardinality": AdjacencyCardinality.describe_yaml_schema(),
                    }
                ],
            }
        )

    @classmethod
    def from_file_data(cls, yaml_data):
        instance = cls()

        for node_data in yaml_data.get("nodes", []):
            node = GraphObjectSchema.from_file_data(node_data)
            instance.put_node_type(node)
        for relationship_data in yaml_data.get("relationships", []):
            relationship = GraphObjectSchema.from_file_data(relationship_data)
            instance.put_relationship_type(relationship)
        for cardinality_data in yaml_data.get("cardinalities", []):
            adjacency = Adjacency.from_file_data(cardinality_data["adjacency"])
            cardinality = AdjacencyCardinality.from_file_data(
                cardinality_data["cardinality"]
            )
            instance.add_adjacency(adjacency, cardinality)
        return instance

    def to_file_data(self):
        return {
            "nodes": [node.to_file_data() for node in self.nodes],
            "relationships": [
                relationship.to_file_data() for relationship in self.relationships
            ],
            "cardinalities": [
                {
                    "adjacency": adjacency.to_file_data(),
                    "cardinality": cardinality.to_file_data(),
                }
                for adjacency, cardinality in self.cardinalities.items()
            ],
        }

    @property
    def nodes_by_name(self) -> Dict[str, GraphObjectSchema]:
        return {node.name: node for node in self.nodes}

    @property
    def relationships_by_name(self) -> Dict[str, GraphObjectSchema]:
        return {relationship.name: relationship for relationship in self.relationships}

    @property
    def nodes(self) -> Iterable[GraphObjectSchema]:
        return (
            schema
            for (type, _), schema in self.type_schemas.items()
            if type == GraphObjectType.NODE
        )

    @property
    def relationships(self) -> Iterable[GraphObjectSchema]:
        return (
            schema
            for (type, _), schema in self.type_schemas.items()
            if type == GraphObjectType.RELATIONSHIP
        )

    @property
    def adjacencies(self) -> Iterable[Adjacency]:
        return self.cardinalities.keys()

    def put_node_type(self, node_type: GraphObjectSchema):
        """Add a node type to the schema.

        Args:
            node_type: The node type to add.

        Returns:
            A new schema with the node type added.
        """
        self.type_schemas[(GraphObjectType.NODE, node_type.name)] = node_type

    def put_relationship_type(self, relationship_type: GraphObjectSchema):
        """Add a relationship type to the schema.

        Args:
            relationship_type: The relationship type to add.

        Returns:
            A new schema with the relationship type added.
        """
        key = (GraphObjectType.RELATIONSHIP, relationship_type.name)
        self.type_schemas[key] = relationship_type

    def drop_node_type_by_name(self, node_type_name: str):
        """Drop a node type from the schema.

        Args:
            node_type_name: The name of the node type to drop.

        Returns:
            A new schema with the node type dropped.
        """
        del self.type_schemas[(GraphObjectType.NODE, node_type_name)]

    def drop_relationship_type_by_name(self, relationship_type_name: str):
        """Drop a relationship type from the schema.

        Args:
            relationship_type_name: The name of the relationship type to drop.

        Returns:
            A new schema with the relationship type dropped.
        """
        del self.type_schemas[(GraphObjectType.RELATIONSHIP, relationship_type_name)]

    def get_by_type_and_object_type(
        self, object_type: GraphObjectType, name: str
    ) -> GraphObjectSchema:
        key = (object_type, name)
        if key not in self.type_schemas:
            self.type_schemas[key] = GraphObjectSchema(name)
        return self.type_schemas[key]

    def get_node_type_by_name(self, node_type_name: str) -> GraphObjectSchema:
        """Get a node type by name.

        If the node type does not exist, a new node type will be created.

        Args:
            node_type_name: The name of the node type to get.

        Returns:
            The node type.
        """
        return self.get_by_type_and_object_type(GraphObjectType.NODE, node_type_name)

    def get_relationship_type_by_name(
        self, relationship_type_name: str
    ) -> GraphObjectSchema:
        """Get a relationship type by name.

        If the relationship type does not exist, a new relationship type will be created.

        Args:
            relationship_type_name: The name of the relationship type to get.

        Returns:
            The relationship type.
        """
        return self.get_by_type_and_object_type(
            GraphObjectType.RELATIONSHIP, relationship_type_name
        )

    def add_adjacency(
        self, adjacency: Adjacency, cardinality: Optional[AdjacencyCardinality] = None
    ):
        """Add an adjacency.

        Args:
            adjacency: The adjacency to add.
        """
        self.cardinalities[adjacency] = cardinality

    def drop_adjacency(self, adjacency: Adjacency):
        """Drop an adjacency.

        Args:
            adjacency: The adjacency to drop.
        """
        del self.cardinalities[adjacency]

    def get_adjacency_cardinality(self, adjacency: Adjacency) -> AdjacencyCardinality:
        """Get an adjacency cardinality.

        Args:
            adjacency: The adjacency.

        Returns:
            The adjacency cardinality.
        """
        return self.cardinalities[adjacency]

    def merge(self, other: "Schema"):
        """Merge another schema into this schema.

        Merges the node types and relationship types from the other schema into
        this schema. If a node type or relationship type already exists in this
        schema, then the node type or relationship type will be merged such that
        the properties of the node type or relationship type in the other schema
        take precedence over the properties of the node type or relationship type
        in this schema.

        Args:
            other: The other schema.
        """
        self.cardinalities.update(other.cardinalities)
        all_types = set(self.type_schemas).union(set(other.type_schemas))
        for obj_type, type in all_types:
            us = self.get_by_type_and_object_type(obj_type, type)
            them = other.get_by_type_and_object_type(obj_type, type)
            us.merge(them)

    def has_node_of_type(self, node_type_name: str) -> bool:
        """Check if a node type exists in the schema.

        Args:
            node_type_name: The name of the node type.

        Returns:
            True if the node type exists in the schema.
        """
        return (GraphObjectType.NODE, node_type_name) in self.type_schemas

    def has_relationship_of_type(self, relationship_type_name: str) -> bool:
        """Check if a relationship type exists in the schema.

        Args:
            relationship_type_name: The name of the relationship type.

        Returns:
            True if the relationship type exists in the schema.
        """
        return (
            GraphObjectType.RELATIONSHIP,
            relationship_type_name,
        ) in self.type_schemas

    def diff_node_types(self, other: "Schema") -> Tuple[Set[str], Set[str]]:
        """Diff node types with another schema.

        Args:
            other: The other schema.

        Returns:
            A tuple of sets. The first set contains the node types that are in
            this schema but not in the other schema. The second set contains
            the node types that are in the other schema but not in this schema.
        """
        us = {node.name for node in self.nodes}
        them = {node.name for node in other.nodes}
        return us - them, them - us

    def diff_relationship_types(self, other: "Schema") -> Tuple[Set[str], Set[str]]:
        """Diff relationship types with another schema.

        Args:
            other: The other schema.

        Returns:
            A tuple of sets. The first set contains the relationship types that
            are in this schema but not in the other schema. The second set
            contains the relationship types that are in the other schema but not
            in this schema.
        """
        us = {relationship.name for relationship in self.relationships}
        them = {relationship.name for relationship in other.relationships}
        return us - them, them - us


@dataclass(slots=True, frozen=True)
class UnboundAdjacency:
    """An unbound adjacency."""

    from_type_or_alias: str
    to_type_or_alias: str
    relationship_type: str
    from_cardinality: Cardinality
    to_cardinality: Cardinality

    def bind(self, schema: Schema, aliases: LayeredDict[str, str]):
        from_type = aliases.get(self.from_type_or_alias, self.from_type_or_alias)
        to_type = aliases.get(self.to_type_or_alias, self.to_type_or_alias)
        schema.add_adjacency(
            Adjacency(from_type, to_type, self.relationship_type),
            AdjacencyCardinality(self.from_cardinality, self.to_cardinality),
        )


@dataclass(slots=True, frozen=True)
class SchemaExpansionCoordinator:
    """A coordinator for expanding a schema."""

    schema: Schema
    aliases: LayeredDict[str, str] = field(default_factory=LayeredDict)
    unbound_aliases: LayeredDict[str, GraphObjectSchema] = field(
        default_factory=LayeredDict
    )
    unbound_properties: LayeredDict[str, GraphObjectSchema] = field(
        default_factory=LayeredDict
    )
    unbound_adjacencies: LayeredList[UnboundAdjacency] = field(
        default_factory=LayeredList
    )

    def on_node_schema(
        self,
        fn: Callable[[GraphObjectSchema], None],
        node_type: Optional[str] = None,
        property_list: Optional[list[str]] = None,
        alias: Optional[str] = None,
    ):
        """Calls a Function on each type of node in the schema aliased.

        Depending on the context, there may be zero or more node schemas that
        match the given alias (if provided). If no alias is provided, then the
        function will be called on just the schema for the node type name.

        If an alias is provided, then the function will be called on all node
        schemas that match the alias.

        Args:
            fn: The function to call.
            node_type_name: The node type name.
            alias: The alias.
        """
        # If both the node_type_name and alias are provided, we are "concreting"
        # the node type name. This means that we are binding the alias(if it exists)
        # to the node type name.
        if node_type and alias:
            node_schema = self.schema.get_node_type_by_name(node_type)
            unbound = self.unbound_aliases.get(alias, GraphObjectSchema(node_type))
            unbound.name = node_type
            node_schema.merge(unbound)
            self.aliases[alias] = node_type
            fn(node_schema)

        # If both the property_list and alias are provided, we are adding properties to the existing
        # object for the alias. We also save this in the unbound aliases to be contextualized.
        elif property_list and alias:
            unbound = self.unbound_aliases.get(alias, GraphObjectSchema(alias))
            unbound.properties.update(
                {property_name: PropertyMetadata() for property_name in property_list}
            )
            self.unbound_aliases[alias] = unbound
            fn(unbound)

        # If only the node_type_name is provided, we are not messing with the alias at all.
        # We are just calling the function on the node type name.
        elif node_type:
            node_schema = self.schema.get_node_type_by_name(node_type)
            fn(node_schema)

    def on_relationship_schema(
        self,
        fn: Callable[[GraphObjectSchema], None],
        relationship_type: str,
    ):
        """Calls a Function to modify the specified relationship schema."""
        fn(self.schema.get_relationship_type_by_name(relationship_type))

    def connect(
        self,
        from_type_or_alias: str,
        to_type_or_alias: str,
        relationship_type: str,
        from_cardinality: Cardinality,
        to_cardinality: Cardinality,
    ):
        """Connect two node types via a relationship type.

        When connecting two node types, the node types may be aliased. If the
        node types are aliased, then the aliases will be used to connect the
        node types by looking up the real node type names. If the node types
        are not aliased, then the node type names will be used directly.

        Args:
            from_type_or_alias: The from node type or alias.
            to_type_or_alias: The to node type or alias.
            relationship_type: The relationship type.
            from_cardinality: The from cardinality.
            to_cardinality: The to cardinality.
        """
        unbound = UnboundAdjacency(
            from_type_or_alias,
            to_type_or_alias,
            relationship_type,
            from_cardinality,
            to_cardinality,
        )
        self.unbound_adjacencies.append(unbound)

    """
        For child expanders, this method is called to maintain the context layer. 
    """

    @contextmanager
    def aquire_context(self, should_be_distinct: bool):
        if should_be_distinct:
            try:
                self.increment_context_level()
                yield
            finally:
                self.decrement_context_level()
        else:
            try:
                yield
            finally:
                pass

    def increment_context_level(self):
        self.unbound_adjacencies.increment_context_level()
        self.unbound_aliases.increment_context_level()
        self.aliases.increment_context_level()

    def decrement_context_level(self):
        self.unbound_adjacencies.decrement_context_level()
        self.unbound_aliases.decrement_context_level()
        self.aliases.decrement_context_level()

    def clear_aliases(self):
        for unbound_adjacency in self.unbound_adjacencies:
            unbound_adjacency.bind(self.schema, self.aliases)


class ExpandsSchema:
    """An interface for an object that expands a schema."""

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        """Expand a schema.

        Args:
            schema: The schema to expand.

        Returns:
            The expanded schema.
        """
        pass

    def make_schema(self) -> Schema:
        """Generates a new schema.

        Returns:
            The new schema.
        """
        coordinator = SchemaExpansionCoordinator(schema := Schema())
        self.expand_schema(coordinator)
        return schema


class ExpandsSchemaFromChildren(ExpandsSchema, ABC):
    """An interface for an object that expands a schema from its children."""

    @abstractmethod
    def get_child_expanders(self) -> Iterable[ExpandsSchema]:
        """Expand a schema from its children.

        Args:
            schema: The schema to expand.

        Returns:
            The expanded schema.
        """
        pass

    """
        In most cases, we do not want to inherit the context between children when we expand the schema.
    """

    @property
    def should_be_distinct(self):
        return True

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        """Expand a schema.

        Args:
            schema: The schema to expand.

        Returns:
            The expanded schema.
        """
        for child_expander in self.get_child_expanders():
            with coordinator.aquire_context(self.should_be_distinct):
                child_expander.expand_schema(coordinator)



from ...schema import Schema
from ...schema.migrations import Migration, MigratorInput, ProjectMigrations
from .operation import NodestreamCommand, Operation


class CleoMigrationInput(MigratorInput):
    def __init__(self, command: NodestreamCommand):
        self.command = command

    def ask_yes_no(self, question: str) -> bool:
        return self.command.confirm(question, default=False)


class GenerateMigration(Operation):
    def __init__(
        self,
        migrations: ProjectMigrations,
        current_state: Schema,
        dry_run: bool = False,
    ):
        self.current_state = current_state
        self.migrations = migrations
        self.dry_run = dry_run

    def describe_migration(self, command: NodestreamCommand, migration: Migration):
        num_changes = len(migration.operations)
        command.line(f"Generated migration {migration.name}.")
        command.line(f"The migration contains {num_changes} schema changes.")

        for operation in migration.operations:
            command.line(f"  - {operation.describe()}")

    async def generate_migration(self, input: CleoMigrationInput):
        migration = path = None

        # If we are dry running, just get the changes as a migration.
        # If we are doing this for real, we can get the path and the migration.
        if self.dry_run:
            migration = await self.migrations.detect_changes(input, self.current_state)
        else:
            result = self.migrations.create_migration_from_changes(
                input, self.current_state
            )
            if (result := await result) is not None:
                migration, path = result

        return migration, path

    async def perform(self, command: NodestreamCommand):
        input = CleoMigrationInput(command)
        migration, path = await self.generate_migration(input)

        # Describe any changes that were made if they were made.
        if migration is not None:
            self.describe_migration(command, migration)
        else:
            command.line("No changes to migrate")

        # Output the path if the migration was commmited to disk.
        if path is not None:
            command.line(f"Migration written to {path}")
            command.line("Run `nodestream migrations run` to apply the migration.")
